from enum import Enum

from mesido.potential_errors import MesidoAssetIssueType, get_potential_errors

from strenum import StrEnum


# Convert potentials errors to errors for the following networks/definitions
class NetworkErrors(StrEnum):
    """
    Enumeration for network type errors.
    """

    HEAT_NETWORK_ERRORS = "heat_network"
    HEAT_AND_COOL_NETWORK_ERRORS = "heat_and_cool_network"
    CUSTOM_ERRORS = "custom_errors"  # an example of custom stuff that can be added in the future
    NO_POTENTIAL_ERRORS_CHECK = "no_potential_errors"


def mesido_issue_type_gen_message(issue_type: MesidoAssetIssueType) -> str:
    """
    Get general message per issue type.

    Returns
    -------
    String message
    """
    type_and_general_meassage = {
        MesidoAssetIssueType.HEAT_PRODUCER_POWER: "Asset insufficient installed capacity: please"
        " increase the installed power or reduce the profile constraint peak value of the"
        " producer(s) listed by choosing a different profile or a different multiplier.",
        MesidoAssetIssueType.HEAT_DEMAND_POWER: "Asset insufficient installed capacity: please"
        " increase the installed power or reduce the demand profile peak value of the demand(s)"
        " listed.",
        MesidoAssetIssueType.COLD_DEMAND_POWER: "Asset insufficient installed capacity: please"
        " increase the installed power or reduce the demand profile peak value of the demand(s)"
        " listed.",
        MesidoAssetIssueType.HEAT_DEMAND_TYPE: "Incorrect asset type: please update.",
        MesidoAssetIssueType.ASSET_PROFILE_CAPABILITY: "Profile assignment not allowed.",
        MesidoAssetIssueType.ASSET_PROFILE_AVAILABILITY: "Profile is not available in the "
        "database.",
        MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_INCORRECT: "Incorrect cost information.",
        MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_MISSING: "Required cost attribute is missing.",
        MesidoAssetIssueType.HEAT_EXCHANGER_TEMPERATURES: "Temperatures at heat exchanger set "
        "incorrectly.",
        MesidoAssetIssueType.HEAT_EXCHANGER_POWER: "The capacity of the heat exchanger is not "
        "defined",
        MesidoAssetIssueType.HEAT_DEMAND_STATE: "Heating Demand state set to OPTIONAL",
        MesidoAssetIssueType.ASSET_PROFILE_MULTIPLIER: "Incorrect asset profile multiplier",
        MesidoAssetIssueType.COLD_ASSET_TYPE_NOT_SUPPORTED: "Unsupported assets are being used.",
        MesidoAssetIssueType.ELECT_ASSET_TYPE_NOT_SUPPORTED: "Unsupported assets are being used.",
    }

    return type_and_general_meassage[issue_type]


def potential_error_to_error(network_check_type: Enum) -> None:
    """
    Convert potential errors to errors for the define error types

    """

    errors_on_types = {
        NetworkErrors.HEAT_NETWORK_ERRORS: [
            MesidoAssetIssueType.HEAT_PRODUCER_POWER,
            MesidoAssetIssueType.HEAT_DEMAND_POWER,
            MesidoAssetIssueType.COLD_DEMAND_POWER,
            MesidoAssetIssueType.HEAT_DEMAND_TYPE,
            MesidoAssetIssueType.ASSET_PROFILE_CAPABILITY,
            MesidoAssetIssueType.ASSET_PROFILE_AVAILABILITY,
            # NOTE: ASSET_COST_ATTRIBUTE_INCORRECT and ASSET_COST_ATTRIBUTE_MISSING
            # are temporarily excluded from HEAT_NETWORK_ERRORS to generate warnings
            # instead of errors, while still performing validation checks
            MesidoAssetIssueType.HEAT_EXCHANGER_TEMPERATURES,
            MesidoAssetIssueType.HEAT_EXCHANGER_POWER,
            MesidoAssetIssueType.HEAT_DEMAND_STATE,
            MesidoAssetIssueType.ASSET_PROFILE_MULTIPLIER,
            MesidoAssetIssueType.COLD_ASSET_TYPE_NOT_SUPPORTED,
            MesidoAssetIssueType.ELECT_ASSET_TYPE_NOT_SUPPORTED,
        ],
        NetworkErrors.HEAT_AND_COOL_NETWORK_ERRORS: [
            MesidoAssetIssueType.HEAT_DEMAND_POWER,
            MesidoAssetIssueType.COLD_DEMAND_POWER,
            MesidoAssetIssueType.HEAT_DEMAND_TYPE,
            MesidoAssetIssueType.ASSET_PROFILE_CAPABILITY,
            MesidoAssetIssueType.ASSET_PROFILE_AVAILABILITY,
            MesidoAssetIssueType.HEAT_EXCHANGER_TEMPERATURES,
            MesidoAssetIssueType.HEAT_DEMAND_STATE,
            MesidoAssetIssueType.ASSET_PROFILE_MULTIPLIER,
        ],
        # Example of extra error types / groups that can be added. This one is not used yet.
        NetworkErrors.CUSTOM_ERRORS: [MesidoAssetIssueType.ASSET_PROFILE_CAPABILITY],
        NetworkErrors.NO_POTENTIAL_ERRORS_CHECK: [],
    }

    # Error checking:
    # - installed capacity/power of a heating/cooling demand is sufficient for the specified
    #   demand profile
    # - asset used for a heating demand
    # - profile assignment capability
    # TODO: Once pyhton 3.11 is used, use the following: Group multiple exceptions (as it is
    # possible multiple exceptions at once are true here):
    # https://www.geeksforgeeks.org/exception-groups-in-python/
    for etype in errors_on_types[network_check_type]:
        if get_potential_errors().have_issues_for(etype):
            get_potential_errors().convert_to_exception(etype, mesido_issue_type_gen_message(etype))



import fnmatch
import os
import sys
from pathlib import Path
from unittest import TestCase

import esdl
from esdl.esdl_handler import EnergySystemHandler

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.financial_mixin import calculate_annuity_factor
from mesido.workflows import (
    EndScenarioSizingDiscountedStaged,
    EndScenarioSizingStaged,
    run_end_scenario_sizing,
)

import numpy as np

import pytest

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestUpdatedESDL(TestCase):

    @pytest.mark.post_process
    def test_updated_esdl(self):
        """
        Check that the updated ESDL resulting from the optmizer, is correct by using the PoCTutorial
        and the Grow_workflow. This is done for the actual esdl file and the esdl string created by
        MESIDO. Both these resulting optimized energy systems should be identical and it is only
        the MESIDO esdl saving method that differs.

        Checks:
        - That the esdl saving method (direct ESDL file and ESDL string)
        - That the correct number of KPIs have been added
        - Check that the OPEX costs for an asset with a 15 year lifetime == OPEX over optim time
        horizon
        - That the correct assets have been removed
        - That all the assets have a state=ENABLED
        - The diameter of all the pipes are as expected
        - The aggregation count of the assets, MESIDO problem vs updated ESDL file
        - That the KPI values are represented in the correct units
        - That assets are connected and that the connections per ports were not changed in the
          updated ESDL
        - That the size of the source has been made small. Not checking the exact
        value - not the purpose of these tests
        - The correct number of polygon sub-areas exist
        """

        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.PoCTutorial.src.run_grow_tutorial

        base_folder = (
            Path(examples.PoCTutorial.src.run_grow_tutorial.__file__).resolve().parent.parent
        )
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"

        problem = EndScenarioSizingStaged(
            esdl_file_name="PoC Tutorial.esdl",
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
        )
        problem.pre()

        # Load in optimized esdl in the form of esdl string created by MESIDO
        esh = EnergySystemHandler()
        file = os.path.join(base_folder, "model", "PoC Tutorial_GrowOptimized_esdl_string.esdl")
        optimized_energy_system_esdl_string: esdl.EnergySystem = esh.load_file(file)

        # Load in optimized esdl in the form of the actual optimized esdl file created by MESIDO
        esdl_path = os.path.join(base_folder, "model", "PoC Tutorial_GrowOptimized.esdl")
        optimized_energy_system = problem._ESDLMixin__energy_system_handler.load_file(esdl_path)

        optimized_energy_systems = [optimized_energy_system_esdl_string, optimized_energy_system]

        for energy_system in optimized_energy_systems:
            # Test KPIs in optimized ESDL

            # High level checks of KPIs
            number_of_kpis_top_level_in_esdl = 11
            high_level_kpis_euro = [
                "High level cost breakdown [EUR] (yearly averaged)",
                "High level cost breakdown [EUR] (30.0 year period)",
                "Overall cost breakdown [EUR] (yearly averaged)",
                "Overall cost breakdown [EUR] (30.0 year period)",
                "CAPEX breakdown [EUR] (30.0 year period)",
                "OPEX breakdown [EUR] (yearly averaged)",
                "OPEX breakdown [EUR] (30.0 year period)",
                "Area_76a7: Asset cost breakdown [EUR]",
                "Area_9d0f: Asset cost breakdown [EUR]",
                "Area_a58a: Asset cost breakdown [EUR]",
            ]
            high_level_kpis_wh = [
                "Energy production [Wh] (yearly averaged)",
            ]
            all_high_level_kpis = []
            all_high_level_kpis = high_level_kpis_euro + high_level_kpis_wh

            np.testing.assert_allclose(
                len(energy_system.instance[0].area.KPIs.kpi), number_of_kpis_top_level_in_esdl
            )
            np.testing.assert_allclose(
                len(energy_system.instance[0].area.KPIs.kpi), len(all_high_level_kpis)
            )

            # Assign kpi info that has to be used for compairing optim time horizon vs yearly values
            # kpi_name_list and kpi_label_list should be the same length and in the same order of
            # which the comparison is done
            compare_yearly_lifetime_kpis = {
                # lists of 2 kpis that have to be compared
                "kpi_name_list": [
                    [
                        "High level cost breakdown [EUR] (yearly averaged)",
                        "High level cost breakdown [EUR] (30.0 year period)",
                    ],
                    [
                        "Overall cost breakdown [EUR] (yearly averaged)",
                        "Overall cost breakdown [EUR] (30.0 year period)",
                    ],
                    [
                        "Overall cost breakdown [EUR] (yearly averaged)",
                        "Overall cost breakdown [EUR] (30.0 year period)",
                    ],
                    [
                        "OPEX breakdown [EUR] (yearly averaged)",
                        "OPEX breakdown [EUR] (30.0 year period)",
                    ],
                ],
                # lists of which kpi label has to be compared for kpi_name_list
                "kpi_label_list": [
                    ["OPEX"],
                    ["Variable OPEX"],
                    ["Fixed OPEX"],
                    ["ResidualHeatSource"],
                ],
                "index_high_level_cost_list": [],  # leave this empty, this list length is set below
            }
            for _ in range(len(compare_yearly_lifetime_kpis["kpi_name_list"])):
                compare_yearly_lifetime_kpis["index_high_level_cost_list"].append([])
            if len(compare_yearly_lifetime_kpis["kpi_name_list"]) != len(
                compare_yearly_lifetime_kpis["kpi_label_list"]
            ):
                print("List should be the same length")
                exit(1)

            for ii in range(len(energy_system.instance[0].area.KPIs.kpi)):
                kpi_name = energy_system.instance[0].area.KPIs.kpi[ii].name
                np.testing.assert_array_equal(
                    kpi_name in all_high_level_kpis,
                    True,
                    err_msg=f"KPI name {kpi_name} was not expected in the ESDL",
                )
                if kpi_name in high_level_kpis_euro:
                    np.testing.assert_array_equal(
                        energy_system.instance[0].area.KPIs.kpi[ii].quantityAndUnit.unit.name,
                        "EURO",
                    )
                elif kpi_name in high_level_kpis_wh:
                    np.testing.assert_array_equal(
                        energy_system.instance[0].area.KPIs.kpi[ii].quantityAndUnit.unit.name,
                        "WATTHOUR",
                    )
                else:
                    exit(f"Unexpected KPI name: {kpi_name}")

                # Check optim time horizon vs yearly cost when the lifetime value is
                # 15 years
                for il in range(len(compare_yearly_lifetime_kpis["kpi_name_list"])):

                    if kpi_name in compare_yearly_lifetime_kpis["kpi_name_list"][il]:
                        compare_yearly_lifetime_kpis["index_high_level_cost_list"][il].append(ii)
                        if len(compare_yearly_lifetime_kpis["index_high_level_cost_list"][il]) == 2:
                            for iitem in range(
                                len(
                                    energy_system.instance[0]
                                    .area.KPIs.kpi[ii]
                                    .distribution.stringItem.items
                                )
                            ):
                                if (
                                    energy_system.instance[0]
                                    .area.KPIs.kpi[
                                        compare_yearly_lifetime_kpis["index_high_level_cost_list"][
                                            il
                                        ][0]
                                    ]
                                    .distribution.stringItem.items[iitem]
                                    .label
                                    in compare_yearly_lifetime_kpis["kpi_label_list"][il]
                                ):

                                    max_value = max(
                                        energy_system.instance[0]
                                        .area.KPIs.kpi[
                                            compare_yearly_lifetime_kpis[
                                                "index_high_level_cost_list"
                                            ][il][0]
                                        ]
                                        .distribution.stringItem.items[iitem]
                                        .value,
                                        energy_system.instance[0]
                                        .area.KPIs.kpi[
                                            compare_yearly_lifetime_kpis[
                                                "index_high_level_cost_list"
                                            ][il][1]
                                        ]
                                        .distribution.stringItem.items[iitem]
                                        .value,
                                    )
                                    min_value = min(
                                        energy_system.instance[0]
                                        .area.KPIs.kpi[
                                            compare_yearly_lifetime_kpis[
                                                "index_high_level_cost_list"
                                            ][il][0]
                                        ]
                                        .distribution.stringItem.items[iitem]
                                        .value,
                                        energy_system.instance[0]
                                        .area.KPIs.kpi[
                                            compare_yearly_lifetime_kpis[
                                                "index_high_level_cost_list"
                                            ][il][1]
                                        ]
                                        .distribution.stringItem.items[iitem]
                                        .value,
                                    )
                                    # Lifetime of 15 years and the optim time horizon is 30 years
                                    np.testing.assert_allclose(min_value * 15.0 * 2.0, max_value)
            # make ssure that all the items in kpi_name_list was checked
            for il in range(len(compare_yearly_lifetime_kpis["kpi_name_list"])):
                np.testing.assert_equal(
                    len(compare_yearly_lifetime_kpis["index_high_level_cost_list"][il]), 2
                )

            # Check the asset quantity
            number_of_assets_in_esdl = 15
            np.testing.assert_allclose(
                len(energy_system.instance[0].area.asset), number_of_assets_in_esdl
            )
            # Check:
            # - that the correct assets were removed
            # - asset state
            # - pipe diameter sizes
            # - asset aggregation count
            # - number of ports
            # - number of connection to a port
            asset_to_be_deleted = ["ResidualHeatSource_76f0", "Pipe_8fa5_ret", "Pipe_8fa5"]
            for ii in range(len(energy_system.instance[0].area.asset)):
                asset_name = energy_system.instance[0].area.asset[ii].name
                # Existance of asset and its state
                np.testing.assert_array_equal(
                    asset_name not in asset_to_be_deleted,
                    True,
                    err_msg=f"Asset name {asset_name} was not expected in the ESDL",
                )
                np.testing.assert_array_equal(
                    energy_system.instance[0].area.asset[ii].state.name == "ENABLED", True
                )

                # Check pipe diameter
                if len(fnmatch.filter([energy_system.instance[0].area.asset[ii].id], "Pipe*")) == 1:
                    if asset_name in ["Pipe1", "Pipe1_ret"]:
                        pipe_size = energy_system.instance[0].area.asset[
                            ii
                        ].diameter == esdl.PipeDiameterEnum.getEEnumLiteral(
                            "DN200"
                        ) or energy_system.instance[
                            0
                        ].area.asset[
                            ii
                        ].diameter == esdl.PipeDiameterEnum.getEEnumLiteral(
                            "DN250"
                        )
                        np.testing.assert_array_equal(pipe_size, True)  # original pipe DN400 being
                        # sized
                    elif asset_name in ["Pipe4", "Pipe4_ret"]:
                        np.testing.assert_array_equal(
                            energy_system.instance[0].area.asset[ii].diameter.name, "DN200"
                        )  # original pipe DN900 being sized
                    elif asset_name not in ["Pipe5", "Pipe5_ret"]:
                        np.testing.assert_array_equal(
                            energy_system.instance[0].area.asset[ii].diameter.name, "DN400"
                        )  # pipe DN was not sized and should be the same as specified in the ESDL
                    else:
                        np.testing.assert_array_equal(
                            energy_system.instance[0].area.asset[ii].diameter.name,
                            "DN300",
                            err_msg=f"Asset name {asset_name} was not expected in the ESDL",
                        )  # pipe DN was not sized and should be the same as specified in the ESDL
                    # Check aggregation count
                    np.testing.assert_array_equal(
                        energy_system.instance[0].area.asset[ii].aggregationCount,
                        problem.get_aggregation_count_max(asset_name),
                    )
                    # Check the number of ports of the assets are as expected
                    np.testing.assert_array_equal(
                        len(energy_system.instance[0].area.asset[ii].port),
                        len(problem.esdl_assets[asset_name].in_ports)
                        + len(problem.esdl_assets[asset_name].out_ports),
                    )
                    # Check the number of connection to a port
                    energy_system.instance[0].area.asset[ii].port[1].name
                    for iport in range(len(energy_system.instance[0].area.asset[ii].port)):
                        if energy_system.instance[0].area.asset[ii].port[iport].name == "In":
                            np.testing.assert_array_equal(
                                len(
                                    energy_system.instance[0]
                                    .area.asset[ii]
                                    .port[iport]
                                    .connectedTo.items
                                ),
                                len(problem.esdl_assets[asset_name].in_ports),
                            )
                    if asset_name == "ResidualHeatSource_72d7":
                        asset_id = energy_system.instance[0].area.asset[ii].id
                        np.testing.assert_array_less(
                            energy_system.instance[0].area.asset[ii].power,
                            problem.esdl_assets[asset_id].attributes["power"],
                        )

            # High level check on the polygon areas drawn
            number_of_areas_in_esdl = 3
            np.testing.assert_allclose(
                len(energy_system.instance[0].area.area), number_of_areas_in_esdl
            )

    @pytest.mark.post_process
    def test_updated_esdl_eac(self):
        """
        Ensure that the updated ESDL, generated through optimization using the
        "discounted_annualized_cost" objective, includes KPIs relevant to the
        Equivalent Annuity Cost (EAC).

        Checks:
        - Confirm that the number of KPIs in the output ESDL matches the expected
            count of KPIs related to EAC.
        - Confirm that the name of KPIs in the output ESDL matches the expected
            name of KPIs related to EAC.
        - Verify that the EAC CAPEX and EAC OPEX values shown in the KPI figures
            are correctly calculated.
        """

        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.PoCTutorial.src.run_grow_tutorial

        base_folder = (
            Path(examples.PoCTutorial.src.run_grow_tutorial.__file__).resolve().parent.parent
        )
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"

        esdl_name = "PoC Tutorial Discount5.esdl"

        # Run the case in order to have access to results and solutions
        solution = run_end_scenario_sizing(
            EndScenarioSizingDiscountedStaged,
            base_folder=base_folder,
            esdl_file_name=esdl_name,
            esdl_parser=ESDLFileParser,
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        problem = EndScenarioSizingDiscountedStaged(
            esdl_file_name=esdl_name,
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
        )
        problem.pre()

        # Load in optimized esdl in the form of the actual optimized esdl file created by MESIDO
        esdl_path = os.path.join(base_folder, "model", f"{esdl_name[:-5]}_GrowOptimized.esdl")
        energy_system = problem._ESDLMixin__energy_system_handler.load_file(esdl_path)

        # Util test
        demand_matching_test(solution, solution.extract_results())
        energy_conservation_test(solution, solution.extract_results())
        heat_to_discharge_test(solution, solution.extract_results())

        # Test KPIs in optimized ESDL
        # High level checks of KPIs
        number_of_kpis_top_level_in_esdl = 8
        high_level_kpis_euro = [
            "EAC - High level cost breakdown [EUR] (1.0 year period)",
            "EAC - Overall cost breakdown [EUR] (1.0 year period)",
            "EAC - CAPEX breakdown [EUR] (1.0 year period)",
            "EAC - OPEX breakdown [EUR] (1.0 year period)",
            "EAC - Area_76a7: Asset cost breakdown [EUR]",
            "EAC - Area_9d0f: Asset cost breakdown [EUR]",
            "EAC - Area_a58a: Asset cost breakdown [EUR]",
        ]
        high_level_kpis_wh = [
            "Energy production [Wh] (yearly averaged)",
        ]
        all_high_level_kpis = []
        all_high_level_kpis = high_level_kpis_euro + high_level_kpis_wh

        np.testing.assert_allclose(
            len(energy_system.instance[0].area.KPIs.kpi), number_of_kpis_top_level_in_esdl
        )
        np.testing.assert_allclose(
            len(energy_system.instance[0].area.KPIs.kpi), len(all_high_level_kpis)
        )

        # Check if EAC calculation is matching with KPI values
        asset = "ResidualHeatSource_72d7"
        for ii in range(len(energy_system.instance[0].area.KPIs.kpi)):
            kpi_name = energy_system.instance[0].area.KPIs.kpi[ii].name

            np.testing.assert_array_equal(
                kpi_name in all_high_level_kpis,
                True,
                err_msg=f"KPI name {kpi_name} was not expected in the ESDL",
            )

            if kpi_name == "EAC - CAPEX breakdown [EUR] (1.0 year period)":
                string_items = (
                    energy_system.instance[0].area.KPIs.kpi[ii].distribution.stringItem.items
                )

                for string_items_asset in string_items:
                    if string_items_asset.label == "ResidualHeatSource":
                        value = string_items_asset.value

                        investment_cost = results[f"{asset}__investment_cost"]
                        installation_cost = results[f"{asset}__installation_cost"]
                        asset_life_years = parameters[f"{asset}.technical_life"]
                        discount_rate = parameters[f"{asset}.discount_rate"] / 100.0
                        annuity_factor = calculate_annuity_factor(discount_rate, asset_life_years)
                        capex_eac = (investment_cost + installation_cost) * annuity_factor

                        np.testing.assert_allclose(value, capex_eac)

            if kpi_name == "EAC - OPEX breakdown [EUR] (1.0 year period)":
                string_items = (
                    energy_system.instance[0].area.KPIs.kpi[ii].distribution.stringItem.items
                )

                for string_items_asset in string_items:
                    if string_items_asset.label == "ResidualHeatSource":
                        value = string_items_asset.value

                        var_opex_cost = results[f"{asset}__variable_operational_cost"]
                        fix_opex_cost = results[f"{asset}__fixed_operational_cost"]

                        opex_eac = var_opex_cost + fix_opex_cost

                        np.testing.assert_allclose(value, opex_eac)


if __name__ == "__main__":
    import time

    start_time = time.time()

    a = TestUpdatedESDL()
    a.test_updated_esdl()
    a.test_updated_esdl_eac()

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from mesido.pycml import Variable
from mesido.pycml.component_library.milp._internal.heat_component import BaseAsset
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from .heat_two_port import HeatTwoPort


@add_variables_documentation_automatically
class _NonStorageComponent(HeatTwoPort, BaseAsset):
    """
    A non storage component is an asset type that does not have states that have interdependent
    time-steps, it does have time-varying states like Heat_flow. In the NonStorageComponent we
    define most of the properties being utilized by all components, like rho and cp.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        # NOTE: We move a factor of 100.0 of the heat to the state entry, to
        # reduce the coefficient in front of the heat variables. This
        # particularly helps the scaling/range of the constraints that relate
        # the milp loss (if it is variable/optional) to the heat in- and out
        # of a component.
        self.Heat_nominal = self.cp * self.rho * self.Q_nominal * 100.0

        self.HeatIn.Heat.nominal = self.Heat_nominal
        self.HeatOut.Heat.nominal = self.Heat_nominal

        self.add_variable(Variable, "Heat_flow", nominal=self.Heat_nominal)



from mesido.pycml.component_library.milp.electricity.electricity_base import ElectricityPort
from mesido.pycml.component_library.milp.heat.heat_source_elec import HeatSourceElec
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan


@add_variables_documentation_automatically
class ElecHeatSourceElec(HeatSourceElec):
    """
    The elec heat source elec component is used to model the source behaviour of e-boilers
    with electricity input port. It functions as a heat source gas that supplies thermal energy
    to the connected heat network. Besides, it models the electricity consumed by the asset.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **modifiers,
        )

        self.component_subtype = "elec_heat_source_elec"

        self.id_mapping_carrier = -1

        self.min_voltage = nan

        # Assumption: heat in/out and added is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(ElectricityPort, "ElectricityIn")

        self.add_equation(
            ((self.ElectricityIn.Power - self.Power_consumed) / self.elec_power_nominal)
        )



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization
from mesido.workflows.rollout_workflow import RollOutProblem

import numpy as np

# from utils_tests import energy_conservation_test
from utils_tests import heat_to_discharge_test


class TestRollOutOptimization(TestCase):

    def test_roll_out_optimization(self):
        """
        Checks:
        - demand matching if placed else demand zero (is_realized variable)
        - heat_to_discharge_test & energy_conservation_test
        - periodiciteit van ATES (end==begin)
        - yearly_storage_initial_value variable implementation (0 when not first timestep of year)
        - yearly max investment
        - to define rollout problem: check that not all heatingdemands are placed in the first
        year, check that all heatingdemands are placed at end of the problem, and both producers
        & ATES.
        - check if integer variables are 0 or 1 for every timestep
        - check number of timesteps in timeseries.
        - if asset placed, also placed for future
        - check fraction is placed
        - check yearly_pipe_length_constraints
        - check if all demands are placed at the end of the simulation
        - check heat buffer constraints


        Missing:
        - Link ATES t0 utilization to state of charge at end of year for optimizations over one
        year.
        """

        # TODO: update location of model also update model & inputfile in run.
        import models.test_case_small_network_ates_buffer_optional_assets.src.run_ates as run_ates

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over three years with timesteps of 30 days
        class RollOutTimeStep(RollOutProblem):
            # pass
            def read(self):
                super().read()
                # m = [3, 5, 5]
                m = [1, 2, 2]
                for i in range(1, 4):
                    demand_timeseries = self.get_timeseries(f"HeatingDemand_{i}.target_heat_demand")
                    demand_timeseries.values[:] = demand_timeseries.values[:] * m[i - 1]
                    self.set_timeseries(f"HeatingDemand_{i}.target_heat_demand", demand_timeseries)

            def solver_options(self):
                options = super().solver_options()
                options["highs"]["mip_rel_gap"] = 0.001
                return options

        solution = run_esdl_mesido_optimization(
            RollOutTimeStep,
            base_folder=base_folder,
            esdl_file_name="PoC_tutorial_incl_ATES_buffer.esdl",
            # esdl_file_name="PoC_tutorial_incl_ATES.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
            yearly_max_capex=6.0e6,
            yearly_max_pipe_length=1.0e3,  # m per year
            include_peak_day=True,
            min_geo_utilization=0.6,
        )
        results = solution.extract_results()

        # tolerance settings for the tests
        atol = 1.0e-1
        rtol = 1.0e-6

        # Demand matching if placed else demand zero (is_realized variable)
        for d in solution.energy_system_components.get("heat_demand", []):
            for y in range(solution._years):
                is_realized = results[f"{d}__asset_is_realized_{y}"]
                if is_realized <= 0.99:
                    # If not placed, demand should be zero
                    np.testing.assert_allclose(
                        results[f"{d}.Heat_demand"][
                            solution._timesteps_per_year
                            * y : solution._timesteps_per_year
                            * (y + 1)
                        ],
                        0.0,
                        atol=atol,
                        rtol=rtol,
                    )
                else:
                    # If placed, demand should match the target heat demand
                    len_times = len(solution.times())

                    target = solution.get_timeseries(f"{d}.target_heat_demand").values[
                        solution._timesteps_per_year * y : len_times
                    ]
                    np.testing.assert_allclose(
                        target,
                        results[f"{d}.Heat_demand"][solution._timesteps_per_year * y : len_times],
                        atol=atol,
                        rtol=rtol,
                    )
                    break  # once an asset is placed it remains placed in the future

        heat_to_discharge_test(solution, results)
        # TODO uncomment once PR 340 is merged and use a slightly relexed tolerance
        # as argument and uncomment next line
        # energy_conservation_test(solution, results)

        assets_to_check = [
            *solution.energy_system_components.get("heat_source", []),
            *solution.energy_system_components.get("heat_demand", []),
            *solution.energy_system_components.get("ates", []),
            *solution.energy_system_components.get("heat_pipe", []),
            *solution.energy_system_components.get("heat_buffer", []),
        ]

        for asset in assets_to_check:
            # Check if integer variables are 0 or 1
            for y in range(solution._years):
                asset_is_realized = results[f"{asset}__asset_is_realized_{y}"]
                np.testing.assert_(
                    np.isclose(asset_is_realized, 0, atol=atol)
                    or np.isclose(asset_is_realized, 1, atol=atol),
                    f"{asset}__asset_is_realized_{y} should be 0 or 1",
                )

            # If asset placed, also placed for future
            for y in range(solution._years - 1):
                asset_is_realized = results[f"{asset}__asset_is_realized_{y}"]
                next_asset_is_realized = results[f"{asset}__asset_is_realized_{y + 1}"]
                tol = 1e-6
                np.testing.assert_(
                    asset_is_realized <= next_asset_is_realized + tol,
                    f"{asset}__asset_is_realized_{y} should be <=\
                        {asset}__asset_is_realized_{y + 1}",
                )

        # Check yearly max investment constraint
        cumulative_prev_year = 0
        for y in range(solution._years):
            cumulative_capex = 0
            cumulative_capex += sum(
                results[f"{asset}__cumulative_investments_made_in_eur_year_{y}"]
                for asset in assets_to_check
            )
            np.testing.assert_(
                cumulative_capex - cumulative_prev_year <= solution._years_timestep_max_capex + tol,
                f"yearly capex {cumulative_capex - cumulative_prev_year}\
                should be <= maximum yearly investment {solution._years_timestep_max_capex}\
                for year {y}",
            )
            cumulative_prev_year = cumulative_capex

        # check number of timesteps in timeseries.
        np.testing.assert_equal(
            len(solution.times()),
            solution._timesteps_per_year * solution._years + 1,
            "Number of timesteps in timeseries is not correct",
        )

        for asset in solution.energy_system_components.get("geothermal", []):
            dt = np.diff(solution.times()) / 3600
            geo_power = solution.bounds()[f"{asset}.Heat_source"][1]
            year_max_power_hour = geo_power * 365 * 24
            fraction_util = solution._min_geo_utilization
            for y in range(solution._years):
                ind_start = y * solution._timesteps_per_year + 1
                ind_end = (y + 1) * solution._timesteps_per_year + 1
                if results[f"{asset}__asset_is_realized_{y}"] >= 0.99:
                    heat_source = results[f"{asset}.Heat_source"]
                    year_prod = sum(
                        heat_source[ind_start:ind_end] * dt[ind_start - 1 : ind_end - 1]
                    )
                    np.testing.assert_(
                        year_prod + atol > fraction_util * year_max_power_hour,
                        f"{asset} is not meeting the minimum operational requirement of "
                        f"{fraction_util*year_max_power_hour}, and produces only {year_prod}.",
                    )

        # Yearly periodicity of ATES
        for ates in solution.energy_system_components.get("ates", []):
            times = solution.times() / 3600 / 24
            for i in range(len(times) - 1):
                if i % solution._timesteps_per_year == 0:
                    np.testing.assert_allclose(
                        results[f"{ates}.Stored_heat"][i],
                        results[f"{ates}.Stored_heat"][i + solution._timesteps_per_year - 1],
                        atol=atol,
                        rtol=rtol,
                    )

        #  Storage_yearly_change variable should be 0 when not first timestep of year
        for ates in solution.energy_system_components.get("ates", []):
            times = solution.times() / 3600 / 24
            for i in range(len(times)):
                if i % solution._timesteps_per_year != 0:
                    np.testing.assert_allclose(
                        results[f"{ates}.Storage_yearly_change"][i],
                        0.0,
                        atol=atol,
                        rtol=rtol,
                    )

        # Check if it is not a trivial roll-out problem, i.e. not all heating demands
        # are placed in the first year
        not_placed_in_first_year = any(
            results[f"{d}__asset_is_realized_0"] <= atol
            for d in solution.energy_system_components.get("heat_demand", [])
        )
        np.testing.assert_(
            not_placed_in_first_year,
            "Trivial roll-out problem: all heating demands are placed in the first year",
        )

        # Check if all heating demands are placed at the end of the problem
        all_placed_at_end = all(
            results[f"{d}__asset_is_realized_{solution._years - 1}"] == 1
            for d in solution.energy_system_components.get("heat_demand", [])
        )
        np.testing.assert_(
            all_placed_at_end,
            "Not all heating demands are placed at the end of the problem",
        )

        # Check if all producers and ATES are placed at the end of the problem
        for asset in [
            *solution.energy_system_components.get("heat_source", []),
            *solution.energy_system_components.get("ates", []),
        ]:
            np.testing.assert_(
                results[f"{asset}__asset_is_realized_{solution._years - 1}"] >= 1 - tol,
                f"Not all assets are placed a the end of the problem {asset} is not placed ("
                f"{results[f'{asset}__asset_is_realized_{solution._years - 1}']}) with solver "
                f"statistics {solution.solver_stats}",
            )

        # Check fraction is placed, should be between 0 and 1 and increasing
        tol = 1e-6
        for asset in assets_to_check:
            for y in range(solution._years):
                asset_fraction_placed = results[f"{asset}__fraction_placed_{y}"]
                np.testing.assert_(
                    0 - tol <= asset_fraction_placed <= 1 + tol, "Value is not between 0 and 1"
                )
                if y < solution._years - 1:
                    next_asset_fraction_placed = results[f"{asset}__fraction_placed_{y + 1}"]
                    np.testing.assert_(
                        asset_fraction_placed <= next_asset_fraction_placed + tol,
                        f"{asset}__fraction_placed_{y} should be <=\
                            {asset}__fraction_placed_{y + 1}",
                    )
            for y in range(solution._years - 1):
                asset_is_realized = results[f"{asset}__asset_is_realized_{y}"]
                next_asset_is_realized = results[f"{asset}__asset_is_realized_{y + 1}"]
                np.testing.assert_(
                    asset_is_realized <= next_asset_is_realized + tol,
                    f"{asset}__asset_is_realized{y} should be <= {asset}__asset_is_realized{y + 1}",
                )

        # Check yearly_pipe_length_constraints
        allowed_length = solution._yearly_max_pipe_length * solution._horizon / solution._years
        total_length = 0
        for y in range(solution._years):
            cumulative_pipe_length = 0
            for p in solution.energy_system_components.get("heat_pipe", []):
                pipe_is_realized = results[f"{p}__asset_is_realized_{y}"]
                pipe_length = solution.parameters(0)[f"{p}.length"]
                cumulative_pipe_length += pipe_is_realized * pipe_length
            used_length_year = cumulative_pipe_length - total_length
            total_length += used_length_year
            print(used_length_year)
            np.testing.assert_(
                used_length_year <= allowed_length,
                f"Yearly pipe length constraint violated in year {y}:\
                    used {used_length_year}, allowed {allowed_length}",
            )

        # Check if all demands placed
        for d in solution.energy_system_components.get("heat_demand", []):
            np.testing.assert_(
                np.isclose(results[f"{d}__asset_is_realized_{solution._years -1}"], 1, atol=atol),
                f"{d} is not placed at the end of the simulation",
            )

        # Heat buffer peak day constraints
        for b in solution.energy_system_components.get("heat_buffer", {}):
            used_buffer = False
            for year in range(solution._years):
                for i in range(solution._timesteps_per_year):
                    if not (
                        i > solution._RollOutProblem__problem_indx_max_peak - 1
                        and i < (solution._RollOutProblem__problem_indx_max_peak + 24 + 1)
                    ):
                        np.testing.assert_allclose(
                            results[f"{b}.Stored_heat"][i + year * solution._timesteps_per_year],
                            0.0,
                            atol=atol,
                            rtol=rtol,
                        )
                        if not i == 0 and not year == 0:
                            np.testing.assert_allclose(
                                results[f"{b}.Heat_buffer"][
                                    i + year * solution._timesteps_per_year
                                ],
                                0.0,
                                atol=atol,
                                rtol=rtol,
                            )
                    else:
                        if (
                            results[f"{b}.Heat_buffer"][i + year * solution._timesteps_per_year]
                            > 1e3
                        ):
                            used_buffer = True

            np.testing.assert_(
                used_buffer, f"{b} Heat buffer has not been used in the entire problem"
            )

        # rollout_post(solution, results)



from mesido.network_common import NetworkSettings
from mesido.pycml.component_library.milp._internal.base_component import BaseAsset


class GasComponent(BaseAsset):
    """
    Base gas component nothing to add here yet.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)
        self.network_type = NetworkSettings.NETWORK_TYPE_GAS



from mesido.pycml import Variable
from mesido.pycml.component_library.milp._internal.heat_component import BaseAsset
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from .heat_two_port import HeatTwoPort


@add_variables_documentation_automatically
class _StorageComponent(HeatTwoPort, BaseAsset):
    """
    A storage component is an asset type that does have states that have interdependent
    time-steps, it does have time-varying states like Heat_flow. In the StorageComponent we
    define most of the properties and variables being utilized by all storage components,
    like heat_nominal (property), heat_flow (variable).

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.minimum_pressure_drop = 1.0e5  # 1 bar of pressure drop
        self.pump_efficiency = 0.5

        self.Heat_nominal = self.cp * self.rho * self.dT * self.Q_nominal

        self.HeatIn.Heat.nominal = self.Heat_nominal
        self.HeatOut.Heat.nominal = self.Heat_nominal

        self.add_variable(Variable, "Heat_flow", nominal=self.Heat_nominal)

        self.add_variable(
            Variable, "Pump_power", min=0.0, nominal=self.Q_nominal * self.nominal_pressure
        )

        self.add_equation(
            (self.Heat_flow - (self.HeatIn.Heat - self.HeatOut.Heat)) / self.Heat_nominal
        )

        self.add_variable(
            Variable,
            "Heat_flow_charging",
            min=0.0,
            max=self.Heat_flow.max,
            nominal=self.Heat_flow.nominal,
        )

        self.add_variable(
            Variable,
            "Heat_flow_discharging",
            min=0.0,
            max=abs(self.Heat_flow.min),
            nominal=self.Heat_flow.nominal,
        )

        self.add_equation(
            (self.Heat_flow - (self.Heat_flow_charging - self.Heat_flow_discharging))
            / self.Heat_nominal
        )



from mesido.pycml import Variable
from mesido.pycml.component_library.milp.electricity.electricity_base import ElectricityPort
from mesido.pycml.component_library.milp.heat.air_water_heat_pump import AirWaterHeatPump
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan


@add_variables_documentation_automatically
class AirWaterHeatPumpElec(AirWaterHeatPump):
    """
    The air-water heat pump component is there to insert thermal power (Heat) into the network.

    The heat to discharge constraints are set in the HeatMixin. We enforce that the outgoing
    temperature of the source matches the absolute thermal power, Q * cp * rho * T_sup == Heat,
    similar as with the demands. This allows us to guarantee that the flow can always carry, as
    the heat losses further downstream in the network are over-estimated with T_ret where in
    reality this temperature drops. It also implicitly assumes that the temperature drops in the
    network are small and thus satisfy minimum temperature requirements.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **modifiers,
        )

        self.component_subtype = "air_water_heat_pump_elec"

        self.id_mapping_carrier = -1

        self.min_voltage = nan
        self.elec_power_nominal = nan

        # Assumption: heat in/out and added is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(ElectricityPort, "ElectricityIn")
        self.add_variable(Variable, "Power_elec", min=0.0, nominal=self.elec_power_nominal)

        self.add_equation(((self.ElectricityIn.Power - self.Power_elec) / self.elec_power_nominal))

        self.add_equation(((self.Power_elec * self.cop - self.Heat_source) / self.Heat_nominal))



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import electric_power_conservation_test

# TODO: still have to make test where elecitricity direction is switched:
# e.g. 2 nodes, with at each node a producer and consumer, first one node medium demand, second
# small demand and then increase the demand of the second node such that direction changes


class TestMILPElectricSourceSink(TestCase):
    def test_source_sink(self):
        """
        Tests for an electricity network that consist out of a source, a cable and a sink.

        Checks:
        - Check that the caps set in the esdl work as intended
        - Check that the consumed power is always>= 0.
        - Check for energy conservation with consumed power, lost power and produced power.
        - Check that the voltage drops over the line.
        - Check the Electricity_source/demand variable is correctly set.
        - Check that minimum voltage is exactly matched.
        - Check that power at the demands equals the current * voltage.

        """

        import models.unit_cases_electricity.source_sink_cable.src.example as example
        from models.unit_cases_electricity.source_sink_cable.src.example import ElectricityProblem

        base_folder = Path(example.__file__).resolve().parent.parent
        tol = 1e-10

        solution = run_esdl_mesido_optimization(
            ElectricityProblem,
            base_folder=base_folder,
            esdl_file_name="case1_elec.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)

        # Test energy conservation
        electric_power_conservation_test(solution, results)

        max_ = solution.bounds()["ElectricityDemand_2af6__max_size"][0]
        v_min = solution.parameters(0)["ElectricityCable_238f.min_voltage"]

        # Test if capping is ok
        power_consumed = results["ElectricityDemand_2af6.ElectricityIn.Power"]
        smallerthen = all(power_consumed <= np.ones(len(power_consumed)) * max_)
        self.assertTrue(smallerthen)
        biggerthen = all(power_consumed >= np.zeros(len(power_consumed)))
        self.assertTrue(biggerthen)

        power_loss = results["ElectricityCable_238f.Power_loss"]
        biggerthen = all(power_loss >= np.zeros(len(power_loss)))
        self.assertTrue(biggerthen)

        # Test that voltage goes down
        v_in = results["ElectricityCable_238f.ElectricityIn.V"]
        v_out = results["ElectricityCable_238f.ElectricityOut.V"]
        np.testing.assert_array_less(v_out, v_in)
        biggerthen = all(v_out >= (v_min - tol) * np.ones(len(v_out)))
        self.assertTrue(biggerthen)

        for source in solution.energy_system_components.get("electricity_source", []):
            np.testing.assert_allclose(
                results[f"{source}.Electricity_source"],
                results[f"{source}.ElectricityOut.Power"],
                atol=1.0e-6,
            )

        for demand in solution.energy_system_components.get("electricity_demand", []):
            np.testing.assert_allclose(
                results[f"{demand}.Electricity_demand"],
                results[f"{demand}.ElectricityIn.Power"],
                atol=1.0e-6,
            )
            np.testing.assert_allclose(
                results[f"{demand}.ElectricityIn.V"],
                parameters[f"{demand}.min_voltage"],
                atol=1.0e-3,
            )
            np.testing.assert_allclose(
                results[f"{demand}.ElectricityIn.V"] * results[f"{demand}.ElectricityIn.I"],
                results[f"{demand}.ElectricityIn.Power"],
                atol=1.0e-3,
            )

    def test_source_sink_max_curr(self):
        """
        Check bounds on the current, this is achieved by increasing the demand forcing the current
        to its max.

        Checks:
        - Check that the caps set in the esdl work as intended
        - Check that the consumed power is always>= 0.
        - Check for energy conservation with consumed power, lost power and produced power.
        - Check that the voltage drops over the line.
        - Check that the current limit is not exceeded
        - Check that minimum voltage is exactly matched
        - Check that power at the demands equals the current * voltage

        """

        import models.unit_cases_electricity.source_sink_cable.src.example as example
        from models.unit_cases_electricity.source_sink_cable.src.example import (
            ElectricityProblemMaxCurr,
        )

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            ElectricityProblemMaxCurr,
            base_folder=base_folder,
            esdl_file_name="case1_elec.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)

        # Test energy conservation
        electric_power_conservation_test(solution, results)

        max_power_transport = (
            parameters["ElectricityCable_238f.min_voltage"]
            * parameters["ElectricityCable_238f.max_current"]
        )  # This max is based on max current and voltage requirement at consumer
        v_min = parameters["ElectricityCable_238f.min_voltage"]  # set as minimum voltage for cables

        tolerance = 1e-10  # due to computational comparison

        # Test if capping is ok (capping based on max power as result of v_min*Imax)
        power_consumed = results["ElectricityDemand_2af6.ElectricityIn.Power"]
        smallerthen = all(
            power_consumed - tolerance <= np.ones(len(power_consumed)) * max_power_transport
        )
        self.assertTrue(smallerthen)
        demand_target = solution.get_timeseries(
            "ElectricityDemand_2af6.target_electricity_demand"
        ).values
        np.testing.assert_allclose(
            power_consumed,
            np.minimum(demand_target, np.ones(len(power_consumed)) * max_power_transport),
        )
        biggerthen = all(power_consumed >= np.zeros(len(power_consumed)))
        self.assertTrue(biggerthen)

        power_loss = results["ElectricityCable_238f.Power_loss"]
        biggerthen = all(power_loss >= np.zeros(len(power_loss)))
        self.assertTrue(biggerthen)

        # Test that voltage goes down
        v_in = results["ElectricityCable_238f.ElectricityIn.V"]
        v_out = results["ElectricityCable_238f.ElectricityOut.V"]
        np.testing.assert_array_less(v_out, v_in)
        biggerthen = all(v_out >= v_min * np.ones(len(v_out)) - tolerance)
        self.assertTrue(biggerthen)

        # Test that max current is not exceeded and is constant along path (since no nodes included)
        current_demand = results["ElectricityDemand_2af6.ElectricityIn.I"]
        current_producer = results["ElectricityProducer_b95d.ElectricityOut.I"]
        current_cable = results["ElectricityCable_238f.ElectricityOut.I"]
        np.testing.assert_allclose(current_demand, current_cable)
        np.testing.assert_allclose(current_cable, current_producer)
        biggerthen = all(
            parameters["ElectricityCable_238f.max_current"] * np.ones(len(current_demand))
            >= current_demand - tolerance
        )
        self.assertTrue(biggerthen)

        for demand in solution.energy_system_components.get("electricity_demand", []):
            np.testing.assert_allclose(
                results[f"{demand}.ElectricityIn.V"],
                parameters[f"{demand}.min_voltage"],
                atol=1.0e-3,
            )
            np.testing.assert_allclose(
                results[f"{demand}.ElectricityIn.V"] * results[f"{demand}.ElectricityIn.I"],
                results[f"{demand}.ElectricityIn.Power"],
                atol=1.0e-3,
            )

    def test_transformer(self):
        """
        This test is to check the transformer component which changes the voltage level.

        Checks:
        - demand matching
        - check the voltage levels are not equal and are correctly set
        - power conservation at the transformer

        """

        import models.unit_cases_electricity.source_sink_cable.src.example as example
        from models.unit_cases_electricity.source_sink_cable.src.example import (
            ElectricityProblem,
        )

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            ElectricityProblem,
            base_folder=base_folder,
            esdl_file_name="transformer.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = solution.extract_results()
        parameters = solution.parameters(0)

        # Check power conservation including power conservation in transformer
        electric_power_conservation_test(solution, results)

        # Check that the cables have two different voltage levels
        assert (
            parameters["ElectricityDemand_2af6.min_voltage"]
            != parameters["Transformer_0185.min_voltage"]
        )

        np.testing.assert_allclose(
            parameters["Transformer_0185.min_voltage"],
            results["ElectricityCable_1fe5.ElectricityOut.V"],
            atol=-1.0e-3,
        )
        np.testing.assert_allclose(
            parameters["ElectricityDemand_2af6.min_voltage"],
            results["ElectricityCable_22c9.ElectricityOut.V"],
            atol=1.0e-3,
        )

        for demand in solution.energy_system_components_get(["electricity_demand", "transformer"]):
            np.testing.assert_allclose(
                results[f"{demand}.ElectricityIn.V"],
                parameters[f"{demand}.min_voltage"],
                atol=1.0e-3,
            )
            np.testing.assert_allclose(
                results[f"{demand}.ElectricityIn.V"] * results[f"{demand}.ElectricityIn.I"],
                results[f"{demand}.ElectricityIn.Power"],
                atol=1.0e-3,
            )



from mesido.pycml import Component

from numpy import nan


class BaseAsset(Component):
    """
    A base asset that carries properties used throughout the different commodities. In this case
    only the financial structure is used in all commodities.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)
        self.state = 1

        self.variable_operational_cost_coefficient = 0.0
        self.fixed_operational_cost_coefficient = 0.0
        self.investment_cost_coefficient = 0.0
        self.installation_cost = 0.0

        self.technical_life = nan
        self.discount_rate = nan
        self.emission_coeff = nan



import logging
import os
import sys
import time

import casadi as ca

from mesido.esdl.esdl_mixin import ESDLMixin
from mesido.head_loss_class import HeadLossOption
from mesido.techno_economic_mixin import TechnoEconomicMixin
from mesido.workflows.goals.rollout_goal import (
    MaximizeRevenueCosts,
    MinimizeCAPEXAssetsCosts,
    MinimizeRolloutFixedOperationalCosts,
    MinimizeVariableOPEX,
)
from mesido.workflows.io.write_output import ScenarioOutput
from mesido.workflows.utils.adapt_profiles import (
    adapt_hourly_profile_averages_timestep_size,
    adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day,
    adapt_profile_for_initial_hour_timestep_size,
    adapt_profile_to_copy_for_number_of_years,
)

import numpy as np

from rtctools._internal.alias_tools import AliasDict
from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.linearized_order_goal_programming_mixin import (
    LinearizedOrderGoalProgrammingMixin,
)
from rtctools.optimization.single_pass_goal_programming_mixin import (
    CachingQPSol,
    SinglePassGoalProgrammingMixin,
)

#  from mesido.workflows.io.rollout_post import rollout_post

logger = logging.getLogger("mesido")
logger.setLevel(logging.INFO)


class SolverHIGHS:
    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "highs"
        highs_options = options["highs"] = {}
        highs_options["mip_rel_gap"] = 0.01

        options["gurobi"] = None
        options["cplex"] = None

        return options


class SolverCPLEX:
    def solver_options(self):
        options = super().solver_options()
        options["casadi_solver"] = self._qpsol
        options["solver"] = "cplex"
        cplex_options = options["cplex"] = {}
        cplex_options["CPX_PARAM_EPGAP"] = 0.01
        options["highs"] = None

        return options


class RollOutProblem(
    # SolverCPLEX,
    SolverHIGHS,
    ScenarioOutput,
    TechnoEconomicMixin,
    LinearizedOrderGoalProgrammingMixin,
    SinglePassGoalProgrammingMixin,
    ESDLMixin,
    CollocatedIntegratedOptimizationProblem,
):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._years = 3
        self._horizon = 30
        self._year_step_size = int(self._horizon / self._years)
        self._timestep_size = kwargs.get("_timestep_size", 30 * 24)

        self._timesteps_per_year = int(365 / (self._timestep_size / 24)) + 1

        # + 1 because of inserting an 1-hour timestep at the beginning of the year,
        # (see adapt_profile_for_initial_hour_timestep_size)
        self._timesteps_per_year += 1

        self._include_peak_day = kwargs.get("include_peak_day", False)
        if self._include_peak_day:
            # + 25 needed if peak day is included with hourly timesteps
            # (see adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day)
            self._timesteps_per_year += 25

        self._yearly_max_capex = kwargs.get("yearly_max_capex", 6.0e6)
        self._yearly_max_pipe_length = kwargs.get("yearly_max_pipe_length", 1.0e3)

        self._years_timestep_max_capex = self._yearly_max_capex * self._horizon / self._years

        # Fraction of how much heat of the total maximum the geo source can produce it should
        # produce in every year that it is placed
        self._min_geo_utilization = kwargs.get("min_geo_utilization", 0.7)

        self._save_json = True

        self.heat_network_settings["head_loss_option"] = HeadLossOption.NO_HEADLOSS
        self.heat_network_settings["minimum_velocity"] = 0.0  # important otherwise heatdemands
        # cannot be turned off for specific timesteps

        self._asset_fraction_placed_map = {}
        self.__asset_fraction_placed_var = {}
        self.__asset_fraction_placed_var_bounds = {}

        # TODO: placeholder for when assets with doublet (integer counts) become available
        self._asset_doublet_is_placed_map = {}
        self.__asset_doublet_is_placed_var = {}
        self.__asset_doublet_is_placed_var_bounds = {}

        self._yearly_capex_var = {}
        self._yearly_capex_var_bounds = {}
        self._yearly_capex_var_nominals = {}

        self._qpsol = None

        self._hot_start = False

        # Store (time taken, success, objective values, solver stats) per priority
        self._priorities_output = []
        self._priority = 0
        self.__priority_timer = None

        self._supported_assets_rollout = [
            "heat_demand",
            "heat_source",
            "heat_pipe",
            "heat_buffer",
            "ates",
        ]

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)
        for pipe in self.energy_system_components.get("heat_pipe", []):
            parameters[f"{pipe}.disconnectable"] = True
        return parameters

    def read(self):
        super().read()

        if self._include_peak_day:
            # Adapt yearly profile with hourly time steps to a common profile (averaged profile
            # except for the day with the peak demand).
            (
                self.__problem_indx_max_peak,
                self.__heat_demand_nominal,
                self.__cold_demand_nominal,
            ) = adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day(
                self,
                self._timestep_size // 24,  # number of days in timestep
            )
        else:
            # Create yearly profile with desired coarser time-step size by
            # averaging over the hourly data
            adapt_hourly_profile_averages_timestep_size(self, self._timestep_size)

        # A small, (1 hour) timestep is inserted as first time step. This is used in the
        # rollout workflow to allow a yearly change in the storage of the ATES system.
        # The first time step is used to accommodate the (yearly) initial storage
        # level of the ATES.
        adapt_profile_for_initial_hour_timestep_size(self)

        if self._include_peak_day:
            # Insertion of the one hour timestep affects self.__problem_indx_max_peak,
            # which is now increased by 1
            self.__problem_indx_max_peak += 1

        # Adapt the profiles to copy for the number of years
        adapt_profile_to_copy_for_number_of_years(self, self._years)

    def pre(self):
        self._qpsol = CachingQPSol()

        super().pre()

        logger.warning(
            "Note: The rollout workflow is still under development and not fully tested yet."
        )

        asset_types = ["low_temperature_ates", "heat_pump"]
        for asset_type in asset_types:
            if len(self.energy_system_components.get(asset_type, [])) > 0:
                logger.error(
                    f"The asset type {asset_type} is not supported in the rollout workflow."
                )
                sys.exit(1)

        # TODO: The asset_fraction_placed is not yet fully functional, eg. not in objective.
        for asset_type in self._supported_assets_rollout:
            for asset in self.energy_system_components.get(asset_type, []):
                self._asset_fraction_placed_map[asset] = []
                for year in range(self._years):
                    asset_fraction_placed_var = f"{asset}__fraction_placed_{year}"
                    self._asset_fraction_placed_map[asset].append(asset_fraction_placed_var)
                    self.__asset_fraction_placed_var[asset_fraction_placed_var] = ca.MX.sym(
                        asset_fraction_placed_var
                    )
                    self.__asset_fraction_placed_var_bounds[asset_fraction_placed_var] = (
                        0.0,
                        1.0,
                    )

        # TODO still needs to be checked if this is also properly added to financial mixing for
        # asset_is_realized for the doublets, then there is no need to do the asset fraction.

        # for asset in [*self.energy_system_components.get("ates", [])]:
        #     for year in range(self._years):
        #         N_doublets = self.parameters(0)[f"{asset}.nr_of_doublets"]
        #         # for i in range(1, N_doublets + 1):
        #         #     asset_is_placed_var = f"{asset}_doublet_{i}__is_placed_{year}"
        #         #     self._asset_doublet_is_placed_map[
        #         #         f"{asset}_doublet_{i}"] = asset_is_placed_var  # might be unnecessary
        #         #     self._asset_is_placed_map[f"{asset}_doublet_{i}"] = asset_is_placed_var
        #         #     self.__asset_doublet_is_placed_var[asset_is_placed_var] = c
        #         #         a.MX.sym(asset_is_placed_var)
        #         #     self.__asset_doublet_is_placed_var_bounds[asset_is_placed_var] = (0.0, 1.0)
        #         #     # self.__asset_is_placed_var[asset_is_placed_var] =
        #         #            ca.MX.sym(asset_is_placed_var)
        #         #     # self.__asset_is_placed_var_bounds[asset_is_placed_var] = (0.0, 1.0)
        #         #
        #         #     asset_fraction_placed_var = f"{asset}_doublet_{i}__fraction_placed_{year}"
        #         #     self._asset_fraction_placed_map[f"{asset}_doublet_{i}"] =
        #         #     asset_fraction_placed_var
        #         #     self.__asset_fraction_placed_var[asset_fraction_placed_var] = \
        #         #         ca.MX.sym(asset_fraction_placed_var)
        #         #     self.__asset_fraction_placed_var_bounds[asset_fraction_placed_var] = (
        #         #     0.0, 1.0)
        #
        #         asset_is_placed_var = f"{asset}__is_placed_{year}"
        #         self._asset_is_placed_map[asset] = asset_is_placed_var
        #         self.__asset_is_placed_var[asset_is_placed_var] = ca.MX.sym(asset_is_placed_var)
        #         self.__asset_is_placed_var_bounds[asset_is_placed_var] = (0.0, 1.0)

        for i in range(self._years):
            var_name = f"yearly_capex_{i}"
            self._yearly_capex_var[var_name] = ca.MX.sym(var_name)
            self._yearly_capex_var_bounds[var_name] = (
                0,
                self._years_timestep_max_capex,
            )
            self._yearly_capex_var_nominals[var_name] = self._years_timestep_max_capex / 2.0

    def energy_system_options(self):
        options = super().energy_system_options()
        options["heat_loss_disconnected_pipe"] = False
        options["neglect_pipe_heat_losses"] = True
        options["include_asset_is_realized"] = True
        options["include_ates_yearly_change_option"] = True
        options["yearly_investments"] = True
        options["min_fraction_tank_volume"] = 0.0
        options["storage_charging_variables"] = False

        return options

    def path_goals(self):
        goals = super().goals().copy()

        goals.append(
            MaximizeRevenueCosts(
                market_price=125e-6,
                year_step_size=self._year_step_size,
                priority=1,
            )
        )

        # goals.append(MinimizeATESState(priority=2))

        return goals

    def goals(self):
        goals = super().goals().copy()

        goals.append(
            MinimizeCAPEXAssetsCosts(
                asset_types_supported=self._supported_assets_rollout, priority=1
            )
        )

        goals.append(
            MinimizeVariableOPEX(
                asset_types_supported=self._supported_assets_rollout,
                year_step_size=self._year_step_size,
                priority=1,
            )
        )

        goals.append(
            MinimizeRolloutFixedOperationalCosts(
                asset_types_supported=self._supported_assets_rollout, priority=1
            )
        )

        return goals

    # TODO: This function might be used for the doublet in later development
    # def __ates_doublet_sums(self, s):
    #     ates_N_doublets = self.parameters(0)[f"{s}.nr_of_doublets"]
    #     ates_doublet_sums = self.get_asset_is__realized_symbols(f"{s}_doublet_{1}")
    #     ates_doublet_fraction_sums = self.get_asset_fraction__placed_symbols(f"{s}_doublet_{1}")
    #     for N in range(1, ates_N_doublets):
    #         ates_doublet_sums += self.get_asset_is__realized_symbols(f"{s}_doublet_{N + 1}")
    #         ates_doublet_fraction_sums += self.get_asset_fraction__placed_symbols(
    #             f"{s}_doublet_{N + 1}"
    #         )

    #     return ates_doublet_sums, ates_doublet_fraction_sums

    def __ates_initial_constraints(self, ensemble_member):
        """
        Initialize ates for first timestep of the simulation.

        """
        constraints = []

        bounds = self.bounds()
        for s in self.energy_system_components.get("ates", []):
            ates_state = self.__state_vector_scaled(f"{s}.Stored_heat", ensemble_member)
            ates_state_big_m = 2.0 * bounds[f"{s}.Stored_heat"][1]

            # For setting the initial state to 0 in the first year the ATES is placed
            constraints.append(((ates_state[0]) / ates_state_big_m, 0.0, 0.0))

            ates_state_big_m = 2.0 * bounds[f"{s}.Heat_ates"][1]
            ates_state = self.__state_vector_scaled(f"{s}.Heat_ates", ensemble_member)
            constraints.append(((ates_state[0]) / ates_state_big_m, 0.0, 0.0))

        return constraints

    def __yearly_asset_is_placed_constraints(self, ensemble_member):
        constraints = []

        for asset, _asset_is_placed_var in self._asset_is_realized_map.items():
            asset_is_placed_vector = self.get_asset_is__realized_symbols(asset)

            constraints.append(
                (
                    (asset_is_placed_vector[1:] - asset_is_placed_vector[:-1]),
                    0.0,
                    np.inf,
                )
            )

        for asset, _asset_fraction_placed_var in self._asset_fraction_placed_map.items():
            asset_fraction_placed_vector = self.get_asset_fraction__placed_symbols(asset)
            constraints.append(
                (
                    (asset_fraction_placed_vector[1:] - asset_fraction_placed_vector[:-1]),
                    0.0,
                    np.inf,
                )
            )

        for (
            asset,
            _asset_fraction_placed_name,
        ) in self._asset_fraction_placed_map.items():
            asset_fraction_placed = self.get_asset_fraction__placed_symbols(asset)
            asset_is_placed_name = self.get_asset_is__realized_symbols(asset)
            constraints.append((asset_is_placed_name - asset_fraction_placed, -np.inf, 0.0))

        return constraints

    def __demand_matching_constraints(self, ensemble_member):
        constraints = []

        # Constraint to enforce matching demands if is_placed
        for d in self.energy_system_components.get("heat_demand", []):
            target = self.get_timeseries(f"{d}.target_heat_demand")

            for year in range(self._years):
                time_start = year * 3600 * 8760
                time_end = (year + 1) * 3600 * 8760
                start_index = np.where(target.times == time_start)[0][0]
                end_index = np.where(target.times == time_end)[0][0]
                demand_states = self.states_in(f"{d}.Heat_demand", time_start, time_end)
                if year == self._years - 1:
                    end_index += 1
                else:
                    demand_states = demand_states[:-1]
                asset_is_realized = self.extra_variable(f"{d}__asset_is_realized_{year}")
                # demand matching
                constraints.append(
                    (
                        (demand_states - asset_is_realized * target.values[start_index:end_index])
                        / target.values[start_index:end_index],
                        0.0,
                        0.0,
                    )
                )

        return constraints

    def __all_demands_placed_constraints(self, ensemble_member):
        """
        Constraint to ensure all heating demands are placed at the end of the optimization

        """
        constraints = []

        for d in self.energy_system_components.get("heat_demand", []):
            asset_realized_vector = self.get_asset_is__realized_symbols(d)

            # ensure asset is placed at end of optimization
            constraints.append((asset_realized_vector[-1], 1.0, 1.0))

        return constraints

    def __yearly_investment_constraints(self, ensemble_member):
        """
        Constraints to set the yearly maximum CAPEX. The CAPEX here is the cumulative investments
        of the assets placed.

        """

        constraints = []

        bounds = self.bounds()

        # Constraint to set yearly maximum CAPEX
        # TODO: CAPEX constraint sources and demands now total capex not yet possible to use
        #  fraction placed variables.
        cumulative_capex_prev_year = 0
        for y in range(self._years):
            cumulative_capex = 0
            for asset_type in self._supported_assets_rollout:
                for asset in self.energy_system_components.get(asset_type, []):
                    # cumulative_investements_made does not yet cather for fraction_placed
                    cumulative_inv = self.extra_variable(
                        f"{asset}__cumulative_investments_made_in_eur_year_{y}"
                    )
                    cumulative_capex += cumulative_inv

            year_nominal = bounds[f"yearly_capex_{y}"][1]
            yearly_capex_var = self.extra_variable(f"yearly_capex_{y}", ensemble_member)
            if y == 0:
                constraints.append(((yearly_capex_var - cumulative_capex) / year_nominal, 0.0, 0.0))
            else:
                constraints.append(
                    (
                        (yearly_capex_var - (cumulative_capex - cumulative_capex_prev_year))
                        / year_nominal,
                        0.0,
                        0.0,
                    )
                )
            cumulative_capex_prev_year = cumulative_capex

        return constraints

    def __yearly_pipe_length_constraints(self, ensemble_member):
        """
        Constraints to set the yearly maximum pipe length that can be placed.
        """
        constraints = []

        parameters = self.parameters(ensemble_member)

        allowed_length = self._yearly_max_pipe_length * self._horizon / self._years
        total_length = 0
        for y in range(self._years):
            cumulative_pipe_length = 0
            for p in self.energy_system_components.get("heat_pipe", []):
                pipe_placement = self.extra_variable(self._asset_is_realized_map[p][y])
                pipe_length = parameters[f"{p}.length"]
                cumulative_pipe_length += pipe_placement * pipe_length
            used_length_year = cumulative_pipe_length - total_length
            total_length += used_length_year

            constraints.append(
                (
                    (allowed_length - used_length_year) / self._yearly_max_pipe_length,
                    0.0,
                    np.inf,
                )
            )

        return constraints

    def __minimum_operational_constraints(self, ensemble_member):
        """
        Constraints to ensure that the geothermal source produces at least a minimum amount of heat
        when it is placed. The minimum amount of heat is defined as a fraction of the maximum
        heat that the geothermal source can produce in a year.

        """

        constraints = []

        bounds = self.bounds()
        for asset in self.energy_system_components.get("geothermal", []):

            logger.warning(
                f"The function {self.__minimum_operational_constraints.__name__} is not tested yet."
            )

            geo_is_placed = self.get_asset_is__realized_symbols(asset)
            heat_produced = self.__state_vector_scaled(f"{asset}.Heat_source", ensemble_member)
            max_heat = bounds[f"{asset}.Heat_source"][1]
            max_heat_year = max_heat * 8760  # 8760 hours in a year

            for year in range(self._years):
                total_heat_year = 0
                dt = np.diff(self.io.times_sec / 3600)
                for i in range(self._timesteps_per_year):
                    total_heat_year += (
                        heat_produced[year * self._timesteps_per_year + i + 1]
                        * dt[year * self._timesteps_per_year + i]
                    )
                constraints.append(
                    (
                        total_heat_year / max_heat_year
                        - geo_is_placed[year] * self._min_geo_utilization,
                        0.0,
                        np.inf,
                    )
                )
        return constraints

    def __ates_yearly_initial_constraints(self, ensemble_member):
        """
        Constraints to set the initial state of an ATES at the first timestep of each year.
        The storage_yearly_change variable is set to zero at all days except the first timestep
        of each year. This variable allow the ATES to have a different initial heat storage levels
        at the beginning of each year.

        """

        constraints = []

        for s in self.energy_system_components.get("ates", []):
            ates_state = self.__state_vector_scaled(f"{s}.Storage_yearly_change", ensemble_member)
            nominal = self.variable_nominal(f"{s}.Heat_ates")
            for i in range(len(self.times())):
                if i % self._timesteps_per_year != 0:
                    # set the storage_yearly_change to zero at all days except first timestep
                    # of each year
                    constraints.append(((ates_state[i]) / nominal, 0.0, 0.0))
        return constraints

    def __ates_yearly_periodic_constraints(self, ensemble_member):
        """
        Constraints to set a yearly periodic condition for an ATES so that the stored
        heat at the end of the year equals that of the beginning of the year.

        """

        constraints = []

        for s in self.energy_system_components.get("ates", []):
            ates_state = self.__state_vector_scaled(f"{s}.Stored_heat", ensemble_member)
            nominal = self.variable_nominal(f"{s}.Stored_heat")
            for i in range(len(self.times()) - 1):
                if i % self._timesteps_per_year == 0:
                    # set the storage first timestep equal to last of each year
                    constraints.append(
                        (
                            (ates_state[i] - ates_state[i + self._timesteps_per_year - 1])
                            / nominal,
                            0.0,
                            0.0,
                        )
                    )
        return constraints

    def __heat_buffer_peak_day_constraints(self, ensemble_member):
        """
        Constraints to ensure that the heat buffer is active only during the 24 hourly timestepping
        period during the peak day.

        """

        constraints = []
        for b in self.energy_system_components.get("heat_buffer", {}):
            var_stored_heat = self.state_vector(f"{b}.Stored_heat")

            for year in range(self._years):
                for i in range(self._timesteps_per_year):
                    if not (
                        i > self.__problem_indx_max_peak - 1
                        and i < (self.__problem_indx_max_peak + 23 + 1)
                    ):
                        constraints.append(
                            (
                                var_stored_heat[i + year * self._timesteps_per_year],
                                0.0,
                                0.0,
                            )
                        )

        return constraints

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)

        constraints.extend(self.__ates_initial_constraints(ensemble_member))
        constraints.extend(self.__demand_matching_constraints(ensemble_member))

        constraints.extend(self.__yearly_asset_is_placed_constraints(ensemble_member))
        constraints.extend(self.__ates_yearly_initial_constraints(ensemble_member))
        constraints.extend(self.__ates_yearly_periodic_constraints(ensemble_member))

        if self._include_peak_day:
            constraints.extend(self.__heat_buffer_peak_day_constraints(ensemble_member))

        constraints.extend(self.__all_demands_placed_constraints(ensemble_member))
        constraints.extend(self.__yearly_investment_constraints(ensemble_member))
        constraints.extend(self.__yearly_pipe_length_constraints(ensemble_member))

        constraints.extend(self.__minimum_operational_constraints(ensemble_member))

        return constraints

    def path_constraints(self, ensemble_member):
        constraints = super().path_constraints(ensemble_member)

        return constraints

    def get_asset_is__realized_symbols(self, asset_name):
        symbols = [f"{asset_name}__asset_is_realized_{year}" for year in range(self._years)]

        return self.extra_variable_vector(symbols, 0)

    def get_asset_fraction__placed_symbols(self, asset_name):
        symbols = [f"{asset_name}__fraction_placed_{year}" for year in range(self._years)]
        return self.extra_variable_vector(symbols, 0)

    def extra_variable_vector(self, symbols, ensemble_member):
        states = []
        for symbol in symbols:
            canonical, sign = self.alias_relation.canonical_signed(symbol)
            nominal = self.variable_nominal(canonical)
            state = nominal * self.state_vector(canonical, ensemble_member)
            states.append(state)
        extra_var_vector = ca.vertcat(*states)
        return extra_var_vector

    def history(self, ensemble_member):
        return AliasDict(self.alias_relation)

    @property
    def extra_variables(self):
        variables = super().extra_variables.copy()
        variables.extend(self._yearly_capex_var.values())
        variables.extend(self.__asset_doublet_is_placed_var.values())
        variables.extend(self.__asset_fraction_placed_var.values())
        return variables

    @property
    def path_variables(self):
        variables = super().path_variables.copy()
        return variables

    def variable_is_discrete(self, variable):
        if (
            # variable in self.__ates_is_charging_var or
            variable
            in self.__asset_doublet_is_placed_var
        ):
            return True
        else:
            return super().variable_is_discrete(variable)

    def variable_nominal(self, variable):
        if variable in self._yearly_capex_var_nominals:
            return self._yearly_capex_var_nominals[variable]
        else:
            return super().variable_nominal(variable)

    def bounds(self):
        bounds = super().bounds()
        bounds.update(self.__asset_doublet_is_placed_var_bounds)
        bounds.update(self._yearly_capex_var_bounds)
        bounds.update(self.__asset_fraction_placed_var_bounds)
        return bounds

    def __state_vector_scaled(self, variable, ensemble_member):
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    def solver_success(self, solver_stats, log_solver_failure_as_error):
        success, log_level = super().solver_success(solver_stats, log_solver_failure_as_error)

        # Allow time-outs for CPLEX and CBC
        if (
            solver_stats["return_status"] == "time limit exceeded"
            or solver_stats["return_status"] == "stopped - on maxnodes, maxsols, maxtime"
        ):

            if self.objective_value > 1e10:
                # Quick check on the objective value. If no solution was
                # found, this is typically something like 1E50.
                return success, log_level

            return True, logging.INFO
        else:
            return success, log_level

    def priority_started(self, priority):
        self._priority = priority
        self.__priority_timer = time.time()

        super().priority_started(priority)

    def priority_completed(self, priority):
        super().priority_completed(priority)
        self._priority = priority
        self._hot_start = True

        time_taken = time.time() - self.__priority_timer
        self._priorities_output.append(
            (
                priority,
                time_taken,
                True,
                self.objective_value,
                self.solver_stats,
            )
        )

    def post(self):
        # In case the solver fails, we do not get in priority_completed(). We
        # append this last priority's statistics here in post().
        success, _ = self.solver_success(self.solver_stats, False)
        if not success:
            time_taken = time.time() - self.__priority_timer
            self._priorities_output.append(
                (
                    self._priority,
                    time_taken,
                    False,
                    self.objective_value,
                    self.solver_stats,
                )
            )

        # Calculate some additional results required/wanted by the CF

        results = self.extract_results()
        parameters = self.parameters(0)

        super().post()

        if os.path.exists(self.output_folder) and self._save_json:
            bounds = self.bounds()
            aliases = self.alias_relation._canonical_variables_map
            solver_stats = self.solver_stats
            self._write_json_output(results, parameters, bounds, aliases, solver_stats)

        # rollout_post(self, results)



import os
import sys
from pathlib import Path
from unittest import TestCase

from mesido.constants import GRAVITATIONAL_CONSTANT
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.util import run_esdl_mesido_optimization

import numpy as np

import pandapipes as pp
from pandapipes.timeseries import run_timeseries

import pandapower.control as control
from pandapower.timeseries import DFData
from pandapower.timeseries import OutputWriter

import pandas as pd

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class ValidateWithPandaPipes(TestCase):
    """
    Test case for a heat network consisting out of a source, pipe(s) and a sink
    """

    def test_heat_network_head_loss(self):
        """
        Heat network: compare the piecewise linear inequality constraints of the head loss
        approximation to pandapipes

        Checks:
        - head losses
        - cp value
        - check pandapipes values
        """
        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.pandapipes.src.run_example
        from examples.pandapipes.src.run_example import HeatProblemHydraulic

        base_folder = Path(examples.pandapipes.src.run_example.__file__).resolve().parent.parent

        class SourcePipeSinkNetwork(HeatProblemHydraulic):

            def energy_system_options(self):
                options = super().energy_system_options()
                self.heat_network_settings["head_loss_option"] = (
                    HeadLossOption.LINEARIZED_N_LINES_WEAK_INEQUALITY
                )
                self.heat_network_settings["n_linearization_lines"] = 10

                self.heat_network_settings["minimum_velocity"] = 0.0

                return options

            def times(self, variable=None) -> np.ndarray:
                """
                Shorten the timeseries to speed up the test

                Parameters
                ----------
                variable : string with name of the variable

                Returns
                -------
                The timeseries
                """
                return super().times(variable)[:10]

        demand_time_series_file = "timeseries_constant.csv"

        solution = run_esdl_mesido_optimization(
            SourcePipeSinkNetwork,
            base_folder=base_folder,
            esdl_file_name="Test_Tree_S1C1.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file=demand_time_series_file,
        )
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)

        # ------------------------------------------------------------------------------
        # ------------------------------------------------------------------------------
        # Setup and run pandapipes
        # ------------------------------------------------------------------------------
        # ------------------------------------------------------------------------------
        from examples.pandapipes.src.pandapipeesdlparser import PandapipeEsdlParserClass

        esdl_file = os.path.join(base_folder, "model", "Test_Tree_S1C1.esdl")
        esdlparser = PandapipeEsdlParserClass()
        esdlparser.loadESDLFile(esdl_file)
        # A producer/consumer cannot directly connect to a pipe, but must be connected to a joint,
        # therefore additional joints created for the inlet & outlet of these assets
        esdlparser.add_additional_joint()

        # ------------------------------------------------------------------------------
        # ------------------------------------------------------------------------------
        # Setup scenario
        total_producers = len(esdlparser.esdl_asset["heat"]["producer"])
        total_consumers = len(esdlparser.esdl_asset["heat"]["consumer"])

        # Create panda_pipes network
        net, net_asset, supply_temperature, return_temperature = esdlparser.createpandapipenet()

        # Setup profile data
        raw_profile_demand_load_watt = pd.read_csv(
            os.path.join(base_folder, "input", demand_time_series_file)
        )
        profile_demand_load_watt = pd.DataFrame(columns=["0"])
        profile_demand_load_watt["0"] = raw_profile_demand_load_watt["demand_1"]

        demand_power = profile_demand_load_watt["0"].to_numpy() / total_consumers
        profile_demand_load_watt = pd.DataFrame(results["demand_1.Heat_demand"])

        # Setup supply mass flow for panda_pipes
        average_temperature_kelvin = (supply_temperature + return_temperature) / 2.0 + 273.15
        cp_joule_kgkelvin = pp.get_fluid(net).get_heat_capacity(average_temperature_kelvin)

        # Enforce mass flow rate instead of cacluting it from Q = m_dot...
        mesido_demand_flow_kg_s = results["Pipe1.Q"] * 988.0
        demand_flow = mesido_demand_flow_kg_s
        supply_flow = demand_flow * total_consumers / total_producers

        # Below control is needed when there is more than 1 producer/heat demand
        net.flow_control.control_active[0] = False  # 1st supplier
        net.flow_control.control_active[0 + total_producers] = False  # 1st demand

        # Assign profiles to the following assets:
        #  - producer mass flow rate, supply temperature
        #  - heat demand power
        # Producer
        isupply = 0
        supply_flow_kg_s = pd.DataFrame(supply_flow, columns=[f"{isupply}"])
        ds_supply_pump_flow_kg_s = DFData(supply_flow_kg_s)
        control.ConstControl(
            net,
            element="circ_pump_mass",
            variable="mdot_flow_kg_per_s",
            element_index=net.circ_pump_mass.index.values[isupply],
            data_source=ds_supply_pump_flow_kg_s,
            profile_name=net.circ_pump_mass.index.values[isupply].astype(str),
        )
        supply_temp_kelvin = pd.DataFrame(
            [supply_temperature + 273.15] * len(profile_demand_load_watt),
            columns=[f"{isupply}"],
        )
        df_supply_pump_temperature_kelvin = DFData(supply_temp_kelvin)
        control.ConstControl(
            net,
            element="circ_pump_mass",
            variable="t_flow_k",
            element_index=net.circ_pump_mass.index.values[isupply],
            data_source=df_supply_pump_temperature_kelvin,
            profile_name=net.circ_pump_mass.index.values[isupply].astype(str),
        )

        # Demand settings
        idemand = 0
        demand_power_watt = pd.DataFrame(demand_power, columns=[f"{idemand}"])
        ds_demand_power_watt = DFData(demand_power_watt)
        control.ConstControl(
            net,
            element="heat_exchanger",
            variable="qext_w",
            element_index=net.heat_exchanger.index.values[idemand],
            data_source=ds_demand_power_watt,
            profile_name=net.heat_exchanger.index.values[idemand].astype(str),
        )

        if profile_demand_load_watt.shape[0] != supply_flow_kg_s.shape[0]:
            exit("profiles do not match")
        # Completed scenario setup

        # ------------------------------------------------------------------------------
        # ------------------------------------------------------------------------------
        # Run panda_pipes simulation
        time_steps = range(profile_demand_load_watt.shape[0])
        log_variables = [
            ("res_pipe", "v_mean_m_per_s"),
            ("res_pipe", "p_from_bar"),
            ("res_pipe", "p_to_bar"),
            ("res_pipe", "mdot_from_kg_per_s"),
            ("heat_exchanger", "qext_w"),
        ]
        ow = OutputWriter(
            net,
            time_steps,
            output_path=None,
            log_variables=log_variables,
        )

        try:
            run_timeseries(net, time_steps, mode="all", friction_model="colebrook")
        except Exception as e:
            print(e)
            exit("Pandapipes runs was not successful")

        # ------------------------------------------------------------------------------
        # Post processing panda_pipes results and tests
        net = esdlparser.correcting_pressure_return(net)
        ow.np_results["heat_exchanger.qext_w"]
        mdata_points = len(mesido_demand_flow_kg_s)

        density = (
            ow.np_results["res_pipe.mdot_from_kg_per_s"][0:mdata_points]
            / ow.np_results["res_pipe.v_mean_m_per_s"][0:mdata_points]
            / (np.pi * net.pipe.diameter_m[0] * net.pipe.diameter_m[0] / 4.0)
        )
        pandapipes_head_loss_m = (
            (
                ow.np_results["res_pipe.p_to_bar"][0:mdata_points]
                - ow.np_results["res_pipe.p_from_bar"][0:mdata_points]
            )
            * 100.0e3
            / density[0][0]
            / GRAVITATIONAL_CONSTANT
        )

        # Compare head losses
        # Hard coded value of 9 used below, since the last data entry has value close to 0, and
        # a comparison is not of importance
        for ii in range(len(results["Pipe1.dH"][:9])):
            np.testing.assert_array_less(
                pandapipes_head_loss_m[ii][0], 0.0
            )  # check that values are negative
            # check that mesido > pandapipes within %
            np.testing.assert_array_less(
                results["Pipe1.dH"][ii] / pandapipes_head_loss_m[ii][0], 1.08
            )
            np.testing.assert_array_less(
                1.0, results["Pipe1.dH"][ii] / pandapipes_head_loss_m[ii][0]
            )
        # Check cp value
        np.testing.assert_allclose(4200.0, cp_joule_kgkelvin)

        # Check panpapipes results
        # Expected results (during validation) during validation work
        expected_dh = {  #
            "pandapipes": [
                -1.34309074e01,
                -1.06285745e01,
                -8.15362473e00,
                -6.00606036e00,
                -4.18588773e00,
                -2.69436747e00,
                -1.52897525e00,
                -6.91030646e-01,
                -1.80687567e-01,
                -2.05221922e-05,
            ],
        }
        for ii in range(len(pandapipes_head_loss_m)):
            np.testing.assert_allclose(expected_dh["pandapipes"][ii], pandapipes_head_loss_m[ii][0])

        # ------------------------------------------------------------------------------------------
        # Do not delete the code below.
        # ------------------------------------------------------------------------------------------

        # # Plotting code for manual checking values
        # import matplotlib.pyplot as plt

        # # Setup data for easy plotting
        # velo_m_s = {
        #     "pandapipes": [],
        #     "mesido": results["Pipe1.HeatIn.Q"] / solution.parameters(0)["Pipe1.area"],
        # }
        # head_loss_m = {
        #     "pandapipes": [],
        #     "mesido": -results["Pipe1.dH"],
        # }
        # for ii in range(len(ow.np_results["res_pipe.v_mean_m_per_s"])):
        #     velo_m_s["pandapipes"].append(ow.np_results["res_pipe.v_mean_m_per_s"][ii][0])
        #     head_loss_m["pandapipes"].append(-pandapipes_head_loss_m[ii][0])

        # # Plot data
        # plt.figure().set_size_inches(10, 6)
        # plt.plot(velo_m_s["pandapipes"], head_loss_m["pandapipes"], linewidth=2)
        # plt.plot(
        #     velo_m_s["mesido"],
        #     head_loss_m["mesido"],
        #     "k--",
        #     linewidth=2,
        #     alpha=0.75,
        #     marker="x"
        # )

        # plt.legend(["pandapipes", "MESIDO"], prop={'size': 13})
        # plt.xlabel("Flow velocity [m/s]", fontsize=16)
        # plt.ylabel("Head loss [m]", fontsize=16)
        # plt.grid()
        # plt.xticks(np.linspace(0, 2.5, 6), fontsize=16)
        # plt.yticks(
        #     np.linspace(0, 15.0, 7),
        #     labels=["0.0", "2.5", "5.0", "7.5", "10.0", "12.5", "15.0"],
        #     fontsize=16,
        # )
        # plt.xlim([0., 2.5])
        # plt.tick_params(left=False, right=False, labelleft=False)
        # plt.ylim([0., 15.])
        # plt.tick_params(left=True, right=False, labelleft=True)
        # n_lines = solution.heat_network_settings["n_linearization_lines"]
        # plt.savefig(f"dH MESIDO_{n_lines} vs pandapipes")
        # plt.show()
        # plt.close()

        # end Plotting code for manual checking values
        # ------------------------------------------------------------------------------------------


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = ValidateWithPandaPipes()
    a.test_heat_network_head_loss()

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization
from mesido.workflows.goals.minimize_tco_goal import MinimizeTCO

import numpy as np

import pytest

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestProducerMaxProfile(TestCase):
    """
    A test to verify that the heat produced by a heat producer is capped by a profile, where the
    producer will always produce equal or less than said profile. This constraint is checked for a
    producer, where the producer's profile was also intentionally smaller than required for a
    couple of timesteps

    """

    @pytest.mark.slow_1
    @pytest.mark.timeout(150)  # overrides the default timelimt of the pytest command
    def test_max_producer_scaled_profile(self):
        """
        Use a scaled profile, where the profile was intentionally reduced for a couple of
        time-steps (reducing the profile value at a few time steps). With the producer size not
        being minimized.

        Checks:
        - Standard checks demand matching, energy conservation and heat to discharge
        - Check that the heat_source <= scaled_profile * size_source

        """

        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemProdProfile

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            HeatProblemProdProfile,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)
        tol = 1e-8
        heat_producer = results["GeothermalSource_b702.Heat_source"]
        size_producer = results["GeothermalSource_b702__max_size"]

        heat_producer_profile_scaled = solution.get_timeseries(
            "GeothermalSource_b702.maximum_heat_source"
        ).values
        heat_producer_profile_full = heat_producer_profile_scaled * size_producer

        # check that heat produced is smaller than the profile
        biggerthen = all(heat_producer_profile_full + tol >= heat_producer)
        self.assertTrue(biggerthen)

    @pytest.mark.slow_1
    @pytest.mark.timeout(150)  # overrides the default timelimt of the pytest command
    def test_max_producer_esdl_unscaled_profile(self):
        """
        Use a profile specified in Watts, where the profile was intentionally modified (via the
        profile multiplier) such that it is smaller than the requried heating demand at a couple of
        timesteps. With the producer size being minimized.

        Checks:
        - Standard checks demand matching, energy conservation and heat to discharge
        - Check that the heat source <= specified profile [W]
        - Check that heat produced is capped at a couple of timesteps
        """

        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import BeseProblemProdProfile

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        class HeatProblemESDLProdProfileAdaptedTCO(BeseProblemProdProfile):
            def read(self):
                super().read()
                set_producer = "HeatProducer_b702"
                producer_timeseries = self.get_timeseries(
                    f"{set_producer}.maximum_heat_source"
                ).values
                producer_timeseries[0:5] = np.ones(5) * 500.0e3
                self.set_timeseries(f"{set_producer}.maximum_heat_source", producer_timeseries)

            def goals(self):
                goals = super().goals().copy()
                goals.append(MinimizeTCO(priority=20, number_of_years=1))

                return goals

        for problem_class in [BeseProblemProdProfile, HeatProblemESDLProdProfileAdaptedTCO]:

            if problem_class == BeseProblemProdProfile:
                # No sizing of the producer
                esdl_file_used = "3a_esdl_source_unscaled_profile.esdl"
            elif problem_class == HeatProblemESDLProdProfileAdaptedTCO:
                # Sizing of the producer is included
                esdl_file_used = "3a_esdl_source_unscaled_profile_sizing_no_storage.esdl"

            solution = run_esdl_mesido_optimization(
                problem_class,
                base_folder=base_folder,
                esdl_file_name=esdl_file_used,
                esdl_parser=ESDLFileParser,
            )
            results = solution.extract_results()

            demand_matching_test(solution, results)
            energy_conservation_test(solution, results)
            heat_to_discharge_test(solution, results, atol=1.0)
            tol = 1e-4
            heat_produced = results["HeatProducer_b702.Heat_source"]

            if problem_class == BeseProblemProdProfile:
                heat_production_upper_limit = solution.get_timeseries(
                    "HeatProducer_b702.maximum_heat_source"
                ).values
                np.testing.assert_equal(
                    solution.esdl_assets[
                        solution.esdl_asset_name_to_id_map["HeatProducer_b702"]
                    ].attributes["power"],
                    results["HeatProducer_b702__max_size"],
                )
                np.testing.assert_array_less(heat_produced - tol, heat_production_upper_limit)
                np.testing.assert_array_less(
                    3,
                    np.sum(
                        np.isclose(heat_produced, heat_production_upper_limit, atol=tol, rtol=1e-10)
                    ),
                )
                np.testing.assert_array_less(
                    np.sum(
                        np.isclose(heat_produced, heat_production_upper_limit, atol=tol, rtol=1e-10)
                    ),
                    20,
                )  # checking that the upper production limit was not achieved for all entries
            elif problem_class == HeatProblemESDLProdProfileAdaptedTCO:
                heat_production_upper_limit = (
                    solution.get_timeseries("HeatProducer_b702.maximum_heat_source").values
                    / max(solution.get_timeseries("HeatProducer_b702.maximum_heat_source").values)
                    * results["HeatProducer_b702__max_size"]
                )
                np.testing.assert_allclose(
                    max(heat_production_upper_limit),
                    results["HeatProducer_b702__max_size"],
                    atol=1e-9,
                )
                np.testing.assert_array_less(heat_produced - tol, heat_production_upper_limit)
                np.testing.assert_allclose(
                    len(heat_produced) - 5,
                    np.sum(
                        np.isclose(heat_produced, heat_production_upper_limit, atol=tol, rtol=1e-9)
                    ),
                )
            else:
                exit("Problem class not catered for in this test case")

    @pytest.mark.slow_1
    @pytest.mark.timeout(150)  # overrides the default timelimt of the pytest command
    def test_max_producer_esdl_scaled_profile(self):
        """
        Use a scaled profile, where the profile was intentionally reduced for a couple of
        time-steps (reducing the profile value at a few time steps). With the producer size not
        being minimized.

        Checks:
        - Standard checks demand matching, energy conservation and heat to discharge
        - Check that the heat source <= specified profile * producer size
        - Check that heat produced is capped at a couple of timesteps
        """

        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemESDLProdProfile

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            HeatProblemESDLProdProfile,
            base_folder=base_folder,
            esdl_file_name="3a_esdl_source_scaled_profile.esdl",
            esdl_parser=ESDLFileParser,
        )
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)
        tol = 1e-8
        heat_produced = results["HeatProducer_b702.Heat_source"]
        heat_production_upper_limit = (
            solution.get_timeseries("HeatProducer_b702.maximum_heat_source").values
            / max(solution.get_timeseries("HeatProducer_b702.maximum_heat_source").values)
            * results["HeatProducer_b702__max_size"]
        )

        # check that heat produced is smaller than the profile
        np.testing.assert_array_less(heat_produced - tol, heat_production_upper_limit)
        np.testing.assert_equal(
            np.sum(np.isclose(heat_produced, heat_production_upper_limit, atol=tol, rtol=1e-10)), 6
        )


if __name__ == "__main__":

    a = TestProducerMaxProfile()
    # a.test_max_producer_scaled_profile()
    a.test_max_producer_esdl_unscaled_profile()
    # a.test_max_producer_esdl_scaled_profile()



import os

from esdl import esdl
from esdl.esdl_handler import EnergySystemHandler

import matplotlib.patches as mpatches
from matplotlib import pyplot as plt

from mesido.post_processing.post_processing_utils import extract_data_results_alias

import numpy as np


class RollOutPost:
    """
    This class contains the basic functions to plot the results of:
     - the placement of assets geospatially and over time
     - the allocation of assets over time.
    """

    def __init__(self, **kwargs):
        self.output_folder = kwargs.get("output_folder")
        self.input_folder = kwargs.get("input_folder", self.output_folder)
        self.model_folder = kwargs.get("model_folder", self.output_folder)
        self.figure_folder = kwargs.get("figure_folder")
        self.esdl_file_name = kwargs.get("esdl_file_name", "ESDL_file.esdl")

        self.results, raw_data = extract_data_results_alias(self.output_folder)
        self.parameters = raw_data["parameters"]
        self.bounds = raw_data["bounds"]

        self.times = np.asarray(self.parameters["times"])
        self.years = int((self.times[-1] - self.times[0]) / 8760 / 3600)
        self._horizon = 30  # total years that are represented by the self.years

        self.esh = EnergySystemHandler()
        self.esh.load_file(os.path.join(self.model_folder, self.esdl_file_name))

    def all_plots(self):
        """
        This function runs all plotting function available for rollout and saves them to the
        output folder.
        """

        self.plot_asset_allocation()
        self.plot_geograph_time()
        self.plot_financials()

    def plot_asset_allocation(self):
        """
        This function plots the allocation of assets (demands, pipes, storages, sources)
        over time and is saved the output folder.
        """
        results = self.results
        ates_assets = self.esh.get_all_instances_of_type(esdl.ATES)

        figure, ax = plt.subplots()
        times = np.asarray(self.parameters["times"])
        times = times - times[0]
        for ates in ates_assets:
            plt.plot(
                times / 3600 / 24, results[f"{ates.name}.Stored_heat"] / 1e9, label=str(ates.name)
            )

        plt.xlabel("Time [days]")
        plt.ylabel("Stored Heat [GJ]")
        plt.title("Heat Storage vs Time")
        plt.legend()
        plt.xticks(range(0, int(times[-1] / 3600 / 24) + 1, 20), minor=True)
        coarse_ticks = [0, 365, 730, 1095]
        plt.xticks(coarse_ticks, [str(t) for t in coarse_ticks])
        plt.grid()
        plt.tight_layout()
        savefig = os.path.join(self.figure_folder, "heat_storage_vs_time.png")
        plt.savefig(savefig)
        plt.close()

        figure, ax = plt.subplots()
        heat_sources = self.esh.get_all_instances_of_type(esdl.HeatProducer)
        for heatsource in heat_sources:
            plt.plot(
                times / 3600 / 24,
                results[f"{heatsource.name}.Heat_source"] / 1e6,
                label=str(heatsource),
            )

        plt.xlabel("Time [days]")
        plt.ylabel("Heat produced [MW]")
        plt.title("Heat [produced] vs Time")
        plt.legend()
        plt.xticks(range(0, int(times[-1] / 3600 / 24) + 1, 20), minor=True)
        coarse_ticks = [0, 365, 730, 1095]
        plt.xticks(coarse_ticks, [str(t) for t in coarse_ticks])
        plt.grid()
        plt.tight_layout()
        savefig = os.path.join(self.figure_folder, "heat_produced_vs_time.png")
        plt.savefig(savefig)
        plt.close()

        figure, ax = plt.subplots()
        for ates_asset in ates_assets:
            ates = ates_asset.name
            plt.plot(
                times / 3600 / 24,
                np.array(results[f"{ates}.Heat_ates"]) / 1e6,
                label=str(ates) + " Heat_ates",
            )
            plt.plot(
                times / 3600 / 24,
                np.array(results[f"{ates}.Heat_loss"]) / 1e6,
                label=str(ates) + " Heat_loss",
            )
            plt.plot(
                times / 3600 / 24,
                np.array(results[f"{ates}.Storage_yearly_change"]) / 1e6,
                label=str(ates) + " Storage_yearly_change",
            )

        plt.xlabel("Time [days]")
        plt.ylabel("Heat [MW]")
        plt.title("Heat  vs Time")
        plt.legend()
        plt.xticks(range(0, int(times[-1] / 3600 / 24) + 1, 20), minor=True)
        coarse_ticks = [0, 365, 730, 1095]
        plt.xticks(coarse_ticks, [str(t) for t in coarse_ticks])
        plt.grid()
        plt.tight_layout()
        savefig = os.path.join(self.figure_folder, "heat_ates_vs_time.png")
        plt.savefig(savefig)
        plt.close()

        total_pipe_length = 0.0
        for pipe in self.esh.get_all_instances_of_type(esdl.Pipe):
            total_pipe_length += pipe.length
        print(total_pipe_length)

    def plot_geograph_time(self):
        """
        This function plots the placements of assets (demands, pipes, storages, sources)
        geographically over time and is saved in the output folder.
        """

        plt.figure()
        color = ["b", "g", "r", "c", "m", "k", "y", "orange", "lime", "teal"]

        legend_years = []
        times = self.times
        problem_years = int((times[-1] - times[0]) / 8760 / 3600)
        yearstep = int(self._horizon / problem_years)
        for i in range(problem_years):
            legend_years.append(
                mpatches.Patch(color=color[i], label=f"year {i * yearstep} {(i + 1) * yearstep}")
            )

        plt.legend(handles=legend_years)
        lat0 = 0.0  # 52.045
        lon0 = 0.0  # 4.315
        for asset in self.esh.get_all_instances_of_type(esdl.EnergyAsset):
            if isinstance(asset, esdl.Pipe):
                name = asset.name
                if name.endswith("_ret"):
                    continue
                line_x = []
                line_y = []
                for point in asset.geometry.point:
                    line_x.append((point.lon - lon0))
                    line_y.append((point.lat - lat0))
                is_placed_list = [
                    self.results[f"{name}__asset_is_realized_{i}"] for i in range(problem_years)
                ]
                if 1 in is_placed_list:
                    idx = np.round(is_placed_list.index(1))
                    plot_size = np.round(asset.innerDiameter * 20)
                    plt.plot(line_x, line_y, color[int(idx)], linewidth=plot_size)
            elif isinstance(asset, esdl.HeatingDemand):
                point = asset.geometry
                line_x = [point.lon - lon0]
                line_y = [point.lat - lat0]
                is_placed_list = [
                    self.results[f"{asset.name}__asset_is_realized_{i}"]
                    for i in range(problem_years)
                ]
                if 1 in is_placed_list:
                    idx = np.round(is_placed_list.index(1))
                    plot_size = np.round(max(self.results[f"{asset.name}.Heat_demand"]) / 1.0e6 * 2)
                    plt.plot(line_x, line_y, "o", color=color[int(idx)], markersize=plot_size)
            elif isinstance(asset, esdl.HeatProducer):
                point = asset.geometry
                line_x = [point.lon - lon0]
                line_y = [point.lat - lat0]
                is_placed_list = [
                    self.results[f"{asset.name}__asset_is_realized_{i}"]
                    for i in range(problem_years)
                ]
                if 1 in is_placed_list:
                    idx = np.round(is_placed_list.index(1))
                    plot_size = np.round(max(self.results[f"{asset.name}.Heat_flow"]) / 1.0e6)
                    plt.plot(line_x, line_y, "s", color=color[int(idx)], markersize=plot_size)
            elif isinstance(asset, esdl.ATES):
                point = asset.geometry
                line_x = [point.lon - lon0]
                line_y = [point.lat - lat0]
                is_placed_list = [
                    self.results[f"{asset.name}__asset_is_realized_{i}"]
                    for i in range(problem_years)
                ]
                if 1 in is_placed_list:
                    idx = np.round(is_placed_list.index(1))
                    plot_size = np.round(max(self.results[f"{asset.name}.Heat_flow"]) / 1.0e6)
                    plt.plot(line_x, line_y, ">", color=color[int(idx)], markersize=plot_size)
            elif isinstance(asset, esdl.HeatStorage):
                point = asset.geometry
                line_x = [point.lon - lon0]
                line_y = [point.lat - lat0]
                is_placed_list = [
                    self.results[f"{asset.name}__asset_is_realized_{i}"]
                    for i in range(problem_years)
                ]
                if 1 in is_placed_list:
                    idx = np.round(is_placed_list.index(1))
                    plot_size = np.round(max(self.results[f"{asset.name}.Heat_flow"]) / 1.0e6)
                    plt.plot(line_x, line_y, ">", color=color[int(idx)], markersize=plot_size)
            else:
                print(f"Asset {asset.name} is not placed on map")

        savefig = os.path.join(self.figure_folder, "geospatial_time_asset_placement.png")
        plt.savefig(savefig)
        plt.show()

    def plot_financials(self):
        parameters = self.parameters
        bounds = self.bounds
        fixed_opex_type = {}
        variable_opex_type = {}
        capex_cumulative_type = {}
        dt = np.diff(self.times)
        steps_per_year = int(len(self.times) / self.years)
        capex_year_totals = [0 for _ in range(self.years)]
        opex_year_totals = [0 for _ in range(self.years)]
        for asset in self.esh.get_all_instances_of_type(esdl.EnergyAsset):
            if asset in [
                *self.esh.get_all_instances_of_type(esdl.HeatingDemand),
                *self.esh.get_all_instances_of_type(esdl.HeatProducer),
                *self.esh.get_all_instances_of_type(esdl.HeatStorage),
                *self.esh.get_all_instances_of_type(esdl.Pipe),
            ]:
                asset_name = asset.name
                asset_type = str(type(asset))
                is_placed = self.get_asset_is__realized_symbols(asset_name)
                cumulative_capex_var = self.get_cumulative_capex_symbols(asset_name)
                fixed_operational_cost_coeff = parameters[
                    f"{asset_name}.fixed_operational_cost_coefficient"
                ]
                variable_operational_cost_coeff = parameters[
                    f"{asset_name}.variable_operational_cost_coefficient"
                ]
                # fixed_operational_cost =
                # optimization_problem._asset_fixed_operational_cost_map.get(
                #     asset_name)
                heat_flow_var = self.results[f"{asset_name}.Heat_flow"]
                max_power = 0.0
                if not isinstance(asset, esdl.Pipe):
                    max_power = bounds[f"{asset_name}__max_size"][1]
                    if max_power == 0:
                        raise RuntimeError(
                            f"Could not determine the max power of asset {asset_name}"
                        )
                # TODO: this can later be replaced by creating a new variable
                # {asset}_fix_operational_cost_{year} set equal to is_placed[i] *
                # _asset_fixed_operational_cost_map_var using bigm constraints.

                var_opex_costs = heat_flow_var[1:] * variable_operational_cost_coeff * dt / 3600
                if asset_type not in variable_opex_type.keys():
                    variable_opex_type[asset_type] = var_opex_costs
                else:
                    variable_opex_type[asset_type] += var_opex_costs

                for i in range(self.years):
                    if asset_type not in fixed_opex_type.keys():
                        fixed_opex_type[asset_type] = []
                    fixed_opex_costs = is_placed[i] * fixed_operational_cost_coeff * max_power
                    if len(fixed_opex_type[asset_type]) == i + 1:
                        fixed_opex_type[asset_type][i] += fixed_opex_costs
                    else:
                        fixed_opex_type[asset_type].append(fixed_opex_costs)

                    if asset_type not in capex_cumulative_type.keys():
                        capex_cumulative_type[asset_type] = []
                    cumulative_capex_costs = cumulative_capex_var[i]
                    if len(capex_cumulative_type[asset_type]) == i + 1:
                        capex_cumulative_type[asset_type][i] += cumulative_capex_costs
                    else:
                        capex_cumulative_type[asset_type].append(cumulative_capex_costs)

                    capex_year_totals[i] += cumulative_capex_costs
                    opex_year_totals[i] += fixed_opex_costs + sum(
                        var_opex_costs[i * steps_per_year : (i + 1) * steps_per_year]
                    )

        years = [i * self._horizon / self.years for i in range(self.years)]

        capex_year_totals = np.asarray(capex_year_totals)
        opex_year_totals = np.asarray(opex_year_totals)
        cumulative_opex_totals = np.cumsum(opex_year_totals * 30 / self.years)

        plt.figure()
        plt.plot(years, capex_year_totals + cumulative_opex_totals, label="costs")
        # plt.plot(years, revenue, label="revenue all connected")
        plt.plot(years, capex_year_totals, alpha=0.2, label="capex")
        plt.plot(years, cumulative_opex_totals, alpha=0.2, label="cumulative opex")
        plt.xlabel("time [years]")
        plt.ylabel("M€")
        plt.grid()
        plt.title("CAPEX, OPEX and revenue")
        plt.legend()
        savefig = os.path.join(self.figure_folder, "financial.png")
        plt.savefig(savefig)
        plt.show()

    # DO NOT DELETE: potential future use of post processing
    # def animate_init():
    #     line.set_data([], [])
    #     return (line,)
    #
    # def animate_fig(i):
    #     lat0 = 0.0  # 52.045
    #     lon0 = 0.0  # 4.315
    #     for id, asset in esdl_assets.items():
    #         if asset.asset_type == "Pipe":
    #             name = asset.name
    #             if "_ret" in name:
    #                 continue
    #             line_x = []
    #             line_y = []
    #             for point in asset.attributes["geometry"].point:
    #                 line_x.append((point.lon - lon0))
    #                 line_y.append((point.lat - lat0))
    #             is_placed_list = [results[f"{name}__is_placed_{i}"] for i in range(10)]
    #             if 1 in is_placed_list:
    #                 idx = np.round(is_placed_list.index(1))
    #                 if idx == i - 1:
    #                     plot_size = np.round(asset.attributes["innerDiameter"] * 20)
    #                     plt.plot(line_x, line_y, color[int(idx)], linewidth=plot_size)
    #                     # line.set_data(line_x, line_y)#, color[int(idx)], linewidth=plot_size)
    #         if asset.asset_type == "HeatingDemand":
    #             b = asset
    #             point = asset.attributes["geometry"]
    #             line_x = [(point.lon - lon0)]
    #             line_y = [(point.lat - lat0)]
    #             is_placed_list = [results[f"{asset.name}__is_placed_{i}"] for i in range(10)]
    #             if 1 in is_placed_list:
    #                 idx = np.round(is_placed_list.index(1))
    #                 if idx == i - 1:
    #                     plot_size =
    #                     np.round(max(results[f"{asset.name}.Heat_demand"]) / 1.0e6 * 3)
    #                     plt.plot(line_x, line_y, "o", color=color[int(idx)], markersize=plot_size)
    #                     # line.set_data(line_x, line_y)
    #         if (
    #             asset.asset_type == "HeatProducer"
    #             or asset.asset_type == "ResidualHeatSource"
    #             or asset.asset_type == "GeothermalSource"
    #         ):
    #             point = asset.attributes["geometry"]
    #             line_x = [(point.lon - lon0)]
    #             line_y = [(point.lat - lat0)]
    #             is_placed_list = [results[f"{asset.name}__is_placed_{i}"] for i in range(10)]
    #             if 1 in is_placed_list:
    #                 idx = np.round(is_placed_list.index(1))
    #                 if idx == i - 1:
    #                     plt.plot(line_x, line_y, "s", color=color[int(idx)])
    #                     # line.set_data(line_x, line_y)
    #         if asset.asset_type == "ATES":
    #             point = asset.attributes["geometry"]
    #             line_x = [(point.lon - lon0)]
    #             line_y = [(point.lat - lat0)]
    #             is_placed_list = [results[f"{asset.name}__is_placed_{i}"] for i in range(10)]
    #             if 1 in is_placed_list:
    #                 idx = np.round(is_placed_list.index(1))
    #                 if idx == i - 1:
    #                     plt.plot(line_x, line_y, ">", color=color[int(idx)])
    #                     # line.set_data(line_x,line_y)
    #     return (line,)

    def get_asset_is__realized_symbols(self, asset_name):
        var_names = [f"{asset_name}__asset_is_realized_{y}" for y in range(self.years)]
        return [self.results[var_n] for var_n in var_names]

    def get_cumulative_capex_symbols(self, asset_name):
        var_names = [
            f"{asset_name}__cumulative_investments_made_in_eur_year_{y}" for y in range(self.years)
        ]
        return [self.results[var_n] for var_n in var_names]


if __name__ == "__main__":
    kwargs = {
        "output_folder": r"C:\git\mesido\examples\municipality" r"\test2",
        "esdl_file_name": "GROW_withATES_Prod_install_full_GrowOptimized_costsupdated.esdl",
        "figure_folder": r"C:\git\mesido\examples\municipality\test2" r"\figures",
    }
    rolloutpost = RollOutPost(**kwargs)
    rolloutpost.all_plots()



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_mixin import DBAccesType
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestHEX(TestCase):

    def test_heat_exchanger(self):
        """
        Check the modelling of the heat exchanger component which allows two hydraulically
        decoupled networks to exchange heat with each other. It is enforced that heat can only flow
        from the primary side to the secondary side, and heat exchangers are allowed to be disabled
        for timesteps in which they are not used. This is to allow for the temperature constraints
        (T_primary > T_secondary) to become deactivated.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - That the efficiency is correclty implemented for heat from primary to secondary
        - Check that the is_disabled is set correctly.
        - Check if the temperatures provided are physically feasible.

        """
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import (
            HeatProblem,
        )

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent
        # -----------------------------------------------------------------------------------------
        # Do not delete: this is used to manualy check writing out of profile data

        class HeatProblemPost(HeatProblem):
            # def post(self):
            #     super().post()
            #     self._write_updated_esdl(
            #         self._ESDLMixin__energy_system_handler.energy_system,
            #         optimizer_sim=True,
            #     )

            def energy_system_options(self):
                options = super().energy_system_options()
                # self.heat_network_settings["minimize_head_losses"] = True  # used for manual tests
                return options

        # Do not delete kwargs: this is used to manualy check writing out of profile data
        kwargs = {
            "write_result_db_profiles": False,
            "database_connections": [
                {
                    "access_type": DBAccesType.WRITE,
                    "influxdb_host": "localhost",
                    "influxdb_port": 8086,
                    "influxdb_username": None,
                    "influxdb_password": None,
                    "influxdb_ssl": False,
                    "influxdb_verify_ssl": False,
                },
            ],
        }
        # -----------------------------------------------------------------------------------------

        solution = run_esdl_mesido_optimization(
            HeatProblemPost,
            base_folder=base_folder,
            esdl_file_name="heat_exchanger.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
            **kwargs,
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        prim_heat = results["HeatExchange_39ed.Primary_heat"]
        sec_heat = results["HeatExchange_39ed.Secondary_heat"]
        disabled = results["HeatExchange_39ed__disabled"]

        # We check the energy converted betweeen the commodities
        eff = parameters["HeatExchange_39ed.efficiency"]

        demand_matching_test(solution, results)
        heat_to_discharge_test(solution, results)
        energy_conservation_test(solution, results)

        np.testing.assert_allclose(prim_heat * eff, sec_heat)

        # Note that we are not testing the last element as we exploit the last timestep for
        # checking the disabled boolean and the assert statement doesn't work for a difference of
        # zero
        np.testing.assert_allclose(prim_heat[-1], 0.0, atol=1e-5)
        np.testing.assert_allclose(disabled[-1], 1.0)
        np.testing.assert_allclose(disabled[:-1], 0.0)
        # Check that heat is flowing through the hex
        np.testing.assert_array_less(-prim_heat[:-1], 0.0)

        np.testing.assert_array_less(
            parameters["HeatExchange_39ed.Secondary.T_supply"],
            parameters["HeatExchange_39ed.Primary.T_supply"],
        )
        np.testing.assert_array_less(
            parameters["HeatExchange_39ed.Secondary.T_return"],
            parameters["HeatExchange_39ed.Primary.T_return"],
        )

    def test_heat_exchanger_bypass(self):
        """
        Check the modelling of the heat exchanger component which allows two hydraulically
        decoupled networks to exchange heat with each other. It is enforced that heat can only flow
        from the primary side to the secondary side, and heat exchangers are allowed to be disabled
        for timesteps in which they are not used. This is to allow for the temperature constraints
        (T_primary > T_secondary) to become deactivated.
        An option to allow for bypassing of the heat exchanger has been added, such that when the
        heat exchanger is disabled, flow through the heat exchanger is allowed, however no heat
        exchange is allowed, in the case the carriers of both the supply and return on one side
        of the heat exchanger are the same.

        Checks:
        - Standard checks for demand matching and energy conservation.
        - Heat to discharge test is not applied as at one heat exchanger (the bypassed one), the
        heat going out on the primary side will not coincide exactly with the temperature due to
        heatlosses in the network before the heat exchanger.
        - Check that the is_disabled is set correctly.
        - Check if the temperatures provided are physically feasible.
        - Checks that heat exchanger is bypassed, e.g. not exchanging heat, but allowing flow when
        both supply and return on one side have the same temperature.
        """
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import (
            HeatProblem,
        )

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        class HeatProblemByPass(HeatProblem):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.heat_network_settings["heat_exchanger_bypass"] = True

            def energy_system_options(self):
                options = super().energy_system_options()
                options["neglect_pipe_heat_losses"] = False

                return options

        solution = run_esdl_mesido_optimization(
            HeatProblemByPass,
            base_folder=base_folder,
            esdl_file_name="test_hex_bypass.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)

        hex_active = "HeatExchange_e410_copy"
        hex_bypass = "HeatExchange_e410"

        np.testing.assert_allclose(results[f"{hex_active}__disabled"][:-1], 0)
        np.testing.assert_allclose(results[f"{hex_bypass}__disabled"][:-1], 1)

        np.testing.assert_array_less(0.001, results[f"{hex_active}.Primary.Q"][:-1])
        np.testing.assert_array_less(0.001, results[f"{hex_bypass}.Primary.Q"][:-1])

        np.testing.assert_allclose(results[f"{hex_bypass}.Heat_flow"][:-1], 0, atol=1e-9)
        np.testing.assert_array_less(1e5, results[f"{hex_active}.Heat_flow"][:-1])

    def test_heat_exchanger_bypass_varying_temperature(self):
        """
        Check the modelling of the heat exchanger component which allows two hydraulically
        decoupled networks to exchange heat with each other. It is enforced that heat can only flow
        from the primary side to the secondary side, and heat exchangers are allowed to be disabled
        for timesteps in which they are not used. This is to allow for the temperature constraints
        (T_primary > T_secondary) to become deactivated.
        An option to allow for bypassing of the heat exchanger has been added, such that when the
        heat exchanger is disabled, flow through the heat exchanger is allowed, however no heat
        exchange is allowed, in the case the carriers of both the supply and return on one side
        of the heat exchanger are the same.

        Checks:
        - Standard checks for demand matching and energy conservation.
        - Heat to discharge test is not applied as at one heat exchanger (the bypassed one), the
        heat going out on the primary side will not coincide exactly with the temperature due to
        heatlosses in the network before the heat exchanger.
        - Check that the is_disabled is set correctly.
        - Check if the temperatures provided are physically feasible.
        - Checks that heat exchanger is bypassed, e.g. not exchanging heat, but allowing flow when
        both supply and return on one side have the same temperature.
        - Check that temperatures are selected correctly at the heat exchanger
        """

        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import (
            HeatProblem,
        )

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        class HeatProblemByPassMultiTemp(HeatProblem):

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                self.heat_network_settings["heat_exchanger_bypass"] = True

            def energy_system_options(self):
                options = super().energy_system_options()
                options["neglect_pipe_heat_losses"] = False

                return options

            def temperature_carriers(self):
                return self.esdl_carriers

            def temperature_regimes(self, carrier):
                temperatures = []
                if carrier == 829940433102452838:
                    temperatures = [70.0, 65.0, 60.0]  # producer out

                if carrier == 8725433194681736500139:
                    temperatures = [65.0, 60.0]  # first hex out

                return temperatures

            def constraints(self, ensemble_member):
                constraints = super().constraints(ensemble_member)

                carriers = self.temperature_carriers()
                for carrier in carriers.values():
                    carrier_map = carrier["id_number_mapping"]
                    temperature_regimes = self.temperature_regimes(int(carrier_map))
                    if len(temperature_regimes) > 1:
                        carrier_var_name = str(carrier_map) + "_temperature"
                        var_carrier = self.extra_variable(carrier_var_name)
                        for i in range(var_carrier.shape[0] - 1):
                            constraints.append((var_carrier[i] - var_carrier[i + 1], 0.0, 0.0))

                        for temperature in temperature_regimes:
                            selected_temp_vec = self.state_vector(
                                f"{int(carrier_map)}_{temperature}"
                            )
                            for i in range(var_carrier.shape[0] - 1):
                                constraints.append(
                                    (selected_temp_vec[i] - selected_temp_vec[i + 1], 0.0, 0.0)
                                )

                return constraints

            # def solver_options(self):
            #     options = super().solver_options()
            #     options["solver"] = "cplex"
            #     return options

            def times(self, variable=None) -> np.ndarray:
                """
                Shorten the timeseries to speed up the test, when highs is used as a solver.
                """
                return super().times(variable)[:5]

        solution = run_esdl_mesido_optimization(
            HeatProblemByPassMultiTemp,
            base_folder=base_folder,
            esdl_file_name="test_hex_bypass_2.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)

        temp_prod = results["829940433102452838_temperature"]
        temp_hex = results["8725433194681736500139_temperature"]

        # check heat exchanger 1 is bypassed
        hex_active = "HeatExchange_e410_copy"
        hex_bypass = "HeatExchange_e410"

        np.testing.assert_allclose(results[f"{hex_active}__disabled"][:-1], 0)
        np.testing.assert_allclose(results[f"{hex_bypass}__disabled"][:-1], 1)

        np.testing.assert_array_less(0.001, results[f"{hex_active}.Primary.Q"][:-1])
        np.testing.assert_array_less(0.001, results[f"{hex_bypass}.Primary.Q"][:-1])

        np.testing.assert_allclose(results[f"{hex_bypass}.Heat_flow"][:-1], 0, atol=1e-6)
        np.testing.assert_array_less(1e5, results[f"{hex_active}.Heat_flow"][:-1])

        # check lowest temperatures are picked (due to minimum heatloss).
        # check temperatures on bypass side are the same.

        np.testing.assert_allclose(temp_prod, 60.0)
        np.testing.assert_allclose(temp_hex, 60.0)
        np.testing.assert_allclose(temp_hex, temp_prod)


class TestHP(TestCase):
    def test_heat_pump(self):
        """
        Check the modelling of the heat pump component which has a constant COP with no energy loss.
        In this specific problem we expect the use of the secondary source to be maximised as
        electrical heat from the HP is "free".

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - Check that the heat pump is producing according to its COP
        - Check that Secondary source use in minimized
        - Check that the upper bound value for heat producing capacity is the same as specified in
        the esdl


        """
        import models.heatpump.src.run_heat_pump as run_heat_pump
        from models.heatpump.src.run_heat_pump import (
            HeatProblem,
        )

        base_folder = Path(run_heat_pump.__file__).resolve().parent.parent

        # -----------------------------------------------------------------------------------------
        # Do not delete: this is used to manualy check writing out of profile data

        class HeatProblemPost(HeatProblem):
            # def post(self):
            #     super().post()
            #     self._write_updated_esdl(
            #         self._ESDLMixin__energy_system_handler.energy_system,
            #         optimizer_sim=True,
            #     )

            def energy_system_options(self):
                options = super().energy_system_options()
                # self.heat_network_settings["minimize_head_losses"] = True  # used for manual tests
                return options

        # Do not delete kwargs: this is used to manualy check writing out of profile data
        kwargs = {
            "write_result_db_profiles": False,
            "database_connections": [
                {
                    "access_type": DBAccesType.WRITE,
                    "influxdb_host": "localhost",
                    "influxdb_port": 8086,
                    "influxdb_username": None,
                    "influxdb_password": None,
                    "influxdb_ssl": False,
                    "influxdb_verify_ssl": False,
                },
            ],
        }
        # -----------------------------------------------------------------------------------------

        solution = run_esdl_mesido_optimization(
            HeatProblemPost,
            base_folder=base_folder,
            esdl_file_name="heat_pump.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
            **kwargs,
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        prim_heat = results["GenericConversion_3d3f.Primary_heat"]
        sec_heat = results["GenericConversion_3d3f.Secondary_heat"]
        power_elec = results["GenericConversion_3d3f.Power_elec"]

        # Check that only the minimum velocity is flowing through the secondary source.
        cross_sectional_area = parameters["Pipe3.area"]
        np.testing.assert_allclose(
            results["ResidualHeatSource_aec9.Q"] / cross_sectional_area,
            1.0e-3,
            atol=2.5e-5,
        )

        demand_matching_test(solution, results)
        heat_to_discharge_test(solution, results)
        energy_conservation_test(solution, results)

        # We check the energy converted betweeen the commodities
        np.testing.assert_allclose(power_elec * parameters["GenericConversion_3d3f.COP"], sec_heat)
        np.testing.assert_allclose(power_elec + prim_heat, sec_heat)

        # Check that the heat pump upper bound
        for key in solution.esdl_assets.keys():
            if solution.esdl_assets[key].asset_type == "HeatPump":
                np.testing.assert_equal(
                    solution.bounds()["GenericConversion_3d3f.Heat_flow"][1],
                    solution.esdl_assets[key].attributes["power"],
                )


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestHEX()
    a.test_heat_exchanger()

    b = TestHP()
    b.test_heat_pump()



from mesido.pycml import Variable
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from ._non_storage_component_sink_type import _NonStorageComponentSinkType


@add_variables_documentation_automatically
class ColdDemand(_NonStorageComponentSinkType):
    """
    The cold demand component is there to add thermal power (Heat) out of the network. Typically,
    this component is used to model aggregated demands.

    The heat to discharge constraints are set in the HeatPhysicsMixin. We enforce that the outgoing
    temperature of the demand matches the absolute thermal power, Q * cp * rho * T_ret == Heat,
    similar as with the sources. This allows us to guarantee that the flow can always carry the
    heat and that thermal losses are always estimated conservatively, as the heat losses further
    downstream in the network are over-estimated with T_ret where in reality this temperature
    drops. It also implicitly assumes that the temperature drops in the network are small and thus
    satisfy minimum temperature requirements.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.

    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(
                    HeatIn=dict(Heat=dict(min=0.0)),
                    HeatOut=dict(Heat=dict(min=0.0)),
                ),
                modifiers,
            ),
        )

        self.component_type = "cold_demand"

        self.minimum_pressure_drop = 1.0e5  # 1 bar of pressure drop

        # Assumption: heat in/out and extracted is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(Variable, "Cold_demand", min=0.0, nominal=self.Heat_nominal)

        self.add_equation(
            (self.HeatOut.Heat - (self.HeatIn.Heat + self.Cold_demand)) / self.Heat_nominal
        )

        self.add_equation((self.Heat_flow - self.Cold_demand) / self.Heat_nominal)



from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan

from ._non_storage_component_source_type import _NonStorageComponentSourceType


@add_variables_documentation_automatically
class Pump(_NonStorageComponentSourceType):
    """
    The pump component is there to add head to the flow. We assume head can only be added for
    positive flow.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.temperature = nan
        self.carrier_id = -1

        self.component_type = "pump"

        self.pump_efficiency = 0.5

        self.add_equation((self.HeatIn.Heat - self.HeatOut.Heat) / self.Heat_nominal)



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.head_loss_class import HeadLossOption
from mesido.util import run_esdl_mesido_optimization

import numpy as np


class TestGasNetwork(TestCase):
    def test_gas_pipe_top(self):
        """
        This test checks the functioning of topology optimization of gas pipes. It uses
        a symmetrical network where the left side should stay in as those pipes are shorter.

        Checks:
        1. Demand is matched
        2. The correct pipes are removed and remained in place
        3. That the correct producer is used

        """
        import models.gas_pipe_topology.src.example as example
        from models.gas_pipe_topology.src.example import HeatProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        class GasNetworkProblem(HeatProblem):
            def energy_system_options(self):
                options = super().energy_system_options()
                self.gas_network_settings["head_loss_option"] = (
                    HeadLossOption.LINEARIZED_ONE_LINE_EQUALITY
                )
                self.gas_network_settings["minimize_head_losses"] = False
                return options

        solution = run_esdl_mesido_optimization(
            GasNetworkProblem,
            base_folder=base_folder,
            esdl_file_name="2a_gas.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        results = solution.extract_results()

        for demand in solution.energy_system_components.get("gas_demand", []):
            target = solution.get_timeseries(f"{demand}.target_gas_demand").values
            np.testing.assert_allclose(target, results[f"{demand}.Gas_demand_mass_flow"])

        removed_pipes = ["Pipe_a718", "Pipe_9a6f", "Pipe_2927", "Pipe_8592"]
        remained_pipes = ["Pipe_51e4", "Pipe_6b39", "Pipe_f9b0", "Pipe_96bc"]
        for pipe in removed_pipes:
            np.testing.assert_allclose(results[f"{pipe}__gn_diameter"], 0.0, atol=1.0e-6)
            np.testing.assert_allclose(results[f"{pipe}__investment_cost"], 0.0, atol=1.0e-6)
            np.testing.assert_allclose(results[f"{pipe}__gn_max_discharge"], 0.0, atol=1.0e-6)
        for pipe in remained_pipes:
            np.testing.assert_array_less(0.0, results[f"{pipe}__gn_diameter"])
            np.testing.assert_array_less(0.0, results[f"{pipe}__investment_cost"])
            np.testing.assert_equal(True, None is not results[f"{pipe}__gn_max_discharge"])

        np.testing.assert_allclose(results["GasProducer_c92e.GasOut.Q"], 0.0, atol=1e-10)
        np.testing.assert_array_less(0.0, results["GasProducer_17aa.GasOut.Q"])


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestGasNetwork()
    a.test_gas_pipe_top()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))
